# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2003-2006 Gary Bishop.
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
# table for translating virtual keys to X windows key symbols

try:
    set
except NameError:
    from sets import Set as set
    
from pyreadline.unicode_helper import ensure_unicode

validkey =set([u'cancel',      u'backspace',    u'tab',          u'clear',
               u'return',      u'shift_l',      u'control_l',    u'alt_l',
               u'pause',       u'caps_lock',    u'escape',       u'space',
               u'prior',       u'next',         u'end',          u'home',
               u'left',        u'up',           u'right',        u'down',
               u'select',      u'print',        u'execute',      u'snapshot',
               u'insert',      u'delete',       u'help',         u'f1',
               u'f2',          u'f3',           u'f4',           u'f5',
               u'f6',          u'f7',           u'f8',           u'f9',
               u'f10',         u'f11',          u'f12',          u'f13',
               u'f14',         u'f15',          u'f16',          u'f17',
               u'f18',         u'f19',          u'f20',          u'f21',
               u'f22',         u'f23',          u'f24',          u'num_lock',
               u'scroll_lock', u'vk_apps',      u'vk_processkey',u'vk_attn',
               u'vk_crsel',    u'vk_exsel',     u'vk_ereof',     u'vk_play',
               u'vk_zoom',     u'vk_noname',    u'vk_pa1',       u'vk_oem_clear',
               u'numpad0',     u'numpad1',      u'numpad2',      u'numpad3',
               u'numpad4',     u'numpad5',      u'numpad6',      u'numpad7',
               u'numpad8',     u'numpad9',      u'divide',       u'multiply',
               u'add',         u'subtract',     u'vk_decimal'])

escape_sequence_to_special_key = {u"\\e[a" : u"up", u"\\e[b" : u"down", u"del" : u"delete"}

class KeyPress(object):
    def __init__(self, char=u"", shift=False, control=False, meta=False, keyname=u""):
        if control or meta or shift:
            char = char.upper()
        self.info = dict(char=char,
                         shift=shift,
                         control=control,
                         meta=meta,
                         keyname=keyname)
        
    def create(name):
        def get(self):
            return self.info[name]

        def set(self, value):
            self.info[name] = value
        return property(get, set)
    char = create(u"char")
    shift = create(u"shift")
    control = create(u"control")
    meta = create(u"meta")
    keyname = create(u"keyname")
        
    def __repr__(self):
        return u"(%s,%s,%s,%s)"%tuple(map(ensure_unicode, self.tuple()))

    def tuple(self):
        if self.keyname:
            return (self.control, self.meta, self.shift, self.keyname)
        else:
            if self.control or self.meta or self.shift:
                return (self.control, self.meta, self.shift, self.char.upper())
            else:
                return (self.control, self.meta, self.shift, self.char)

    def __eq__(self, other):
        if isinstance(other, KeyPress):
            s = self.tuple()
            o = other.tuple()
            return s == o
        else:
            return False

def make_KeyPress_from_keydescr(keydescr):
    keyinfo = KeyPress()
    if len(keydescr) > 2 and keydescr[:1] == u'"' and keydescr[-1:] == u'"':
        keydescr = keydescr[1:-1]
        
    while 1:
        lkeyname = keydescr.lower()
        if lkeyname.startswith(u'control-'):
            keyinfo.control = True
            keydescr = keydescr[8:]
        elif lkeyname.startswith(u'ctrl-'):
            keyinfo.control = True
            keydescr = keydescr[5:]
        elif keydescr.lower().startswith(u'\\c-'):
            keyinfo.control = True
            keydescr = keydescr[3:]
        elif keydescr.lower().startswith(u'\\m-'):
            keyinfo.meta = True
            keydescr = keydescr[3:]
        elif keydescr in escape_sequence_to_special_key:
            keydescr = escape_sequence_to_special_key[keydescr]
        elif lkeyname.startswith(u'meta-'):
            keyinfo.meta = True
            keydescr = keydescr[5:]
        elif lkeyname.startswith(u'alt-'):
            keyinfo.meta = True
            keydescr = keydescr[4:]
        elif lkeyname.startswith(u'shift-'):
            keyinfo.shift = True
            keydescr = keydescr[6:]
        else:
            if len(keydescr) > 1:
                if keydescr.strip().lower() in validkey:
                    keyinfo.keyname = keydescr.strip().lower()
                    keyinfo.char = ""
                else:
                    raise IndexError(u"Not a valid key: '%s'"%keydescr)
            else:
                keyinfo.char = keydescr
            return keyinfo

if __name__ == u"__main__":
    import startup
    


"""
Makes sure the Variables module for seash works as intended.
In particular, we check that the $varname and $varname$ methods of referencing
variables are functional.
"""

import seash

import seash_dictionary
import seash_modules

#pragma out Enabled modules: modules
seash_modules.enable_modules_from_last_session(seash_dictionary.seashcommanddict)

seash.command_loop([
  'enable variables',
  'set helpcmd help',
  '$helpcmd',
  '$helpcmd show',
  '$helpcmd$',
  '$helpcmd$ show',
  'disable variables',
])


"""
<Program Name>
  modules/command_callbacks.py

<Purpose>
  Defines all the command callbacks that are used for the modules module.

  For more information on how the functions within this module are defined,
  refer to the main command_callbacks.py in the same directory as seash.py.
  
"""
import seash_modules
import seash_dictionary
import seash_helper
import seash_exceptions

# show modules
def list_all_modules(input_dict, environment_dict):
  """
  <Purpose>
    Lists all installed modules and their status.

  <Arguments>
    input_dict:  The commanddict representing the user's input.
    environment_dict:  The dictionary representing the current seash 
                       environment.

  <Side Effects>
    A list of all enabled modules and all installed modules will be printed to 
    stdout.

  <Exceptions>
    None

  <Returns>
    None
  """
  print "Enabled Modules:"
  print ", ".join(seash_modules.get_enabled_modules())
  print
  print "Installed Modules:"

  # Now print the URLs...
  # Output format:
  # modulename - URL not available  # URL is set to None
  # modulename - https://seattle.poly.edu/seashplugins/modulename/ # URL is set
  for module in seash_modules.module_data:
    print module, '-',
    if seash_modules.module_data[module]['url'] is None:
      print "Install URL not available"
    else:
      print seash_modules.module_data[module]['url']
  print


# enable modulename
def enable_module(input_dict, environment_dict):
  """
  <Purpose>
    Enables an installed module.

  <Arguments>
    input_dict:  The commanddict representing the user's input.
    environment_dict:  The dictionary representing the current seash 
                       environment.

  <Side Effects>
    The module specified in the 'modulename' node will be enabled.

  <Exceptions>
    UserError:  input_dict did not contain a 'modulename' node.

  <Returns>
    None
  """

  # Get the modulename
  dict_mark = input_dict
  try:
    command = dict_mark.keys()[0]
    while dict_mark[command]['name'] != 'modulename':
      dict_mark = input_dict[command]['children']
      command = dict_mark.keys()[0]
    modulename = command
  except IndexError:
    raise seash_exceptions.UserError("Error, command requires a modulename")

  try:
    seash_modules.enable(seash_dictionary.seashcommanddict, modulename)
  except seash_exceptions.ModuleConflictError, e:
    print "Module cannot be imported due to the following conflicting command:"
    print str(e)
  except seash_exceptions.InitializeError, e:
    print "Error while enabling the '"+modulename+"' module."
    seash_modules.disable(seash_dictionary.seashcommanddict, modulename)



def disable_module(input_dict, environment_dict):
  """
  <Purpose>
    Disables an enabled module.

  <Arguments>
    input_dict:  The commanddict representing the user's input.
    environment_dict:  The dictionary representing the current seash 
                       environment.

  <Side Effects>
    The module specified in the 'modulename' node will be enabled.

  <Exceptions>
    UserError:
      input_dict did not contain a 'modulename' node, or the user tried to
      disable this module.

  <Returns>
    None
  """

  # Get the modulename
  dict_mark = input_dict
  try:
    command = dict_mark.keys()[0]
    while dict_mark[command]['name'] != 'modulename':
      dict_mark = input_dict[command]['children']
      command = dict_mark.keys()[0]
    modulename = command
  except IndexError:
    raise seash_exceptions.UserError("Error, command requires a modulename")

  if modulename == 'modules':
    raise seash_exceptions.UserError("Error, cannot disable the 'modules' module")

  seash_modules.disable(seash_dictionary.seashcommanddict, modulename)



def print_module_help(input_dict, environment_dict):
  """
  <Purpose>
    Displays the module level help for a module.

  <Arguments>
    input_dict:  The commanddict representing the user's input.
    environment_dict:  The dictionary representing the current seash 
                       environment.

  <Side Effects>
    The helptext for the module specified in the 'modulename' node will be 
    printed to stdout.

  <Exceptions>
    UserError:
      input_dict did not contain a 'modulename' node, or the user specified
      a module that is not installed.

  <Returns>
    None
  """

  # Get the modulename
  dict_mark = input_dict
  try:
    command = dict_mark.keys()[0]
    while dict_mark[command]['name'] != 'modulename':
      dict_mark = input_dict[command]['children']
      command = dict_mark.keys()[0]
    modulename = command
  except IndexError:
    raise seash_exceptions.UserError("Error, command requires a modulename")

  # Is this module installed?
  if not modulename in seash_modules.module_data:
    raise seash_exceptions.UserError("Module is not installed.")

  print seash_modules.module_data[modulename]['help_text']

  # Now, print out all the commands under this module
  print "Commands in this module:"
  print '\n'.join(seash_helper.get_commands_from_commanddict(seash_modules.module_data[modulename]['command_dict']))
  




"""
Author: Alan Loh
Module: Holds and keeps track of the global variables that contains
        information in regards to the current session seash.

Holds any global variables that's not already in environment_dict and
needs to be accessed by multiple files.
"""

# a dict that contains all of the targets (vessels and groups) we know about.
targets = {'%all':[]}

# stores information about the vessels...
vesselinfo = {}

# the nextid that should be used for a new target.
nextid = 1

# a dict that contains all of the key information
keys = {}

# With the addition of affixes, affixes slow down openconnection a lot
# with their lookups and will trigger numerous timeouts.  Make sure
# there's enough time to account for the affixes overhead, at least
# until that is fixed.

# this is how long we wait for a node to timeout
globalseashtimeout = 30

# this is the upload rate we'll modify globalseashtimeout during 
# file uploads to ensure completion without timeout
# default is left at 10240 bytes/sec (10 kb/sec)
globaluploadrate = 5120



# -*- coding: ISO-8859-1 -*-
import re,sys,os

terminal_escape = re.compile(u'(\001?\033\\[[0-9;]*m\002?)')
escape_parts = re.compile(u'\001?\033\\[([0-9;]*)m\002?')


class AnsiState(object):
    def __init__(self,bold=False,inverse=False,color=u"white",background=u"black",backgroundbold=False):
        self.bold = bold
        self.inverse = inverse
        self.color = color
        self.background = background
        self.backgroundbold = backgroundbold

    trtable = {u"black":0, u"red":4,      u"green":2,  u"yellow":6,
               u"blue":1,  u"magenta":5,  u"cyan":3,  u"white":7}
    revtable = dict(zip(trtable.values(),trtable.keys()))
    def get_winattr(self):
        attr = 0
        if self.bold:
            attr |= 0x0008
        if self.backgroundbold:
            attr |= 0x0080
        if self.inverse:
            attr |= 0x4000
        attr |= self.trtable[self.color]
        attr |= (self.trtable[self.background] << 4)
        return attr

    def set_winattr(self, attr):
        self.bold = bool(attr & 0x0008)
        self.backgroundbold = bool(attr & 0x0080)
        self.inverse = bool(attr & 0x4000)
        self.color = self.revtable[attr & 0x0007]
        self.background = self.revtable[(attr & 0x0070) >> 4]
        
    winattr=property(get_winattr,set_winattr)
    def __repr__(self):
        return u'AnsiState(bold=%s,inverse=%s,color=%9s,'    \
               u'background=%9s,backgroundbold=%s)# 0x%x'%   \
                (self.bold, self.inverse, '"%s"'%self.color,  
                 '"%s"'%self.background, self.backgroundbold,
                 self.winattr)

    def copy(self):
        x = AnsiState()
        x.bold = self.bold
        x.inverse = self.inverse
        x.color = self.color
        x.background = self.background
        x.backgroundbold = self.backgroundbold
        return x

defaultstate = AnsiState(False,False,u"white")

trtable = {0:u"black",  1:u"red",       2:u"green", 3:u"yellow",
           4:u"blue",   5:u"magenta",   6:u"cyan",  7:u"white"}

class AnsiWriter(object):
    def __init__(self, default=defaultstate):
        if isinstance(defaultstate, AnsiState):
            self.defaultstate = default
        else:
            self.defaultstate=AnsiState()
            self.defaultstate.winattr = defaultstate
            
            
    def write_color(self,text, attr=None):
        u'''write text at current cursor position and interpret color escapes.

        return the number of characters written.
        '''
        if isinstance(attr,AnsiState):
            defaultstate = attr
        elif attr is None:  #use attribute form initial console
            attr = self.defaultstate.copy()
        else:
            defaultstate = AnsiState()
            defaultstate.winattr = attr
            attr = defaultstate
        chunks = terminal_escape.split(text)
        n = 0 # count the characters we actually write, omitting the escapes
        res=[]
        for chunk in chunks:
            m = escape_parts.match(chunk)
            if m:
                parts = m.group(1).split(u";")
                if len(parts) == 1 and parts[0] == u"0":
                    attr = self.defaultstate.copy()
                    continue
                for part in parts:
                    if part == u"0": # No text attribute
                        attr = self.defaultstate.copy()
                        attr.bold=False
                    elif part == u"7": # switch on reverse
                        attr.inverse=True
                    elif part == u"1": # switch on bold (i.e. intensify foreground color)
                        attr.bold=True 
                    elif len(part) == 2 and u"30" <= part <= u"37": # set foreground color
                        attr.color = trtable[int(part) - 30]
                    elif len(part) == 2 and u"40" <= part <= u"47": # set background color
                        attr.backgroundcolor = trtable[int(part) - 40]
                continue
            n += len(chunk)
            if True:
                res.append((attr.copy(), chunk))
        return n,res

    def parse_color(self,text, attr=None):
        n,res=self.write_color(text, attr)
        return n, [attr.winattr for attr, text in res]

def write_color(text, attr=None):
    a = AnsiWriter(defaultstate)
    return a.write_color(text, attr)

def write_color_old( text, attr=None):
    u'''write text at current cursor position and interpret color escapes.

    return the number of characters written.
    '''
    res = []
    chunks = terminal_escape.split(text)
    n = 0 # count the characters we actually write, omitting the escapes
    if attr is None:#use attribute from initial console
        attr = 15
    for chunk in chunks:
        m = escape_parts.match(chunk)
        if m:
            for part in m.group(1).split(u";"):
                if part == u"0": # No text attribute
                    attr = 0
                elif part == u"7": # switch on reverse
                    attr |= 0x4000
                if part == u"1": # switch on bold (i.e. intensify foreground color)
                    attr |= 0x08
                elif len(part) == 2 and u"30" <= part <= u"37": # set foreground color
                    part = int(part)-30
                    # we have to mirror bits
                    attr = (attr & ~0x07) | ((part & 0x1) << 2) | (part & 0x2) | ((part & 0x4) >> 2)
                elif len(part) == 2 and u"40" <= part <= u"47": # set background color
                    part = int(part) - 40
                    # we have to mirror bits
                    attr = (attr & ~0x70) | ((part & 0x1) << 6) | ((part & 0x2) << 4) | ((part & 0x4) << 2)
                # ignore blink, underline and anything we don't understand
            continue
        n += len(chunk)
        if chunk:
            res.append((u"0x%x"%attr, chunk))
    return res


#trtable={0:"black",1:"red",2:"green",3:"yellow",4:"blue",5:"magenta",6:"cyan",7:"white"}

if __name__==u"__main__x":
    import pprint
    pprint=pprint.pprint

    s=u"\033[0;31mred\033[0;32mgreen\033[0;33myellow\033[0;34mblue\033[0;35mmagenta\033[0;36mcyan\033[0;37mwhite\033[0m"
    pprint (write_color(s))    
    pprint (write_color_old(s))
    s=u"\033[1;31mred\033[1;32mgreen\033[1;33myellow\033[1;34mblue\033[1;35mmagenta\033[1;36mcyan\033[1;37mwhite\033[0m"
    pprint (write_color(s))    
    pprint (write_color_old(s))    

    s=u"\033[0;7;31mred\033[0;7;32mgreen\033[0;7;33myellow\033[0;7;34mblue\033[0;7;35mmagenta\033[0;7;36mcyan\033[0;7;37mwhite\033[0m"
    pprint (write_color(s))    
    pprint (write_color_old(s))
    s=u"\033[1;7;31mred\033[1;7;32mgreen\033[1;7;33myellow\033[1;7;34mblue\033[1;7;35mmagenta\033[1;7;36mcyan\033[1;7;37mwhite\033[0m"
    pprint (write_color(s))    
    pprint (write_color_old(s))    

    
if __name__==u"__main__":
    import console
    import pprint
    pprint=pprint.pprint
    
    c=console.Console()
    c.write_color(u"dhsjdhs")
    c.write_color(u"\033[0;32mIn [\033[1;32m1\033[0;32m]:")
    print
    pprint (write_color(u"\033[0;32mIn [\033[1;32m1\033[0;32m]:"))    
    
if __name__==u"__main__x":
    import pprint
    pprint=pprint.pprint
    s=u"\033[0;31mred\033[0;32mgreen\033[0;33myellow\033[0;34mblue\033[0;35mmagenta\033[0;36mcyan\033[0;37mwhite\033[0m"
    pprint (write_color(s))    



"""
Author: Alan Loh
Module: Holds all non-callback seash-related methods here.

This would include:
-Helper functions
-Functions that operate on a target
"""

###Helper Methods###

import seash_global_variables

import seash_exceptions

from repyportability import *
add_dy_support(locals())

# Use local clock for time if there is no network connectivity
import time

# Used for error checking during file handling
import errno

# Enable the NAT traversal affix
dy_import_module_symbols('affixstackinterface.r2py')

old_openconnection = openconnection
affix_obj = AffixStackInterface('(CoordinationAffix)')

def new_openconnection(destip, destport, localip, localport, timeout):
  if destip.endswith('zenodotus.poly.edu'):
    return affix_obj.openconnection(destip, destport, localip, localport, timeout)
  else:
    return old_openconnection(destip, destport, localip, localport, timeout)

openconnection = new_openconnection

dy_import_module_symbols("time.r2py")

dy_import_module_symbols("rsa.r2py")

dy_import_module_symbols("listops.r2py")

dy_import_module_symbols("parallelize.r2py")

dy_import_module_symbols("domainnameinfo.r2py")

dy_import_module_symbols("advertise.r2py")   #  used to do OpenDHT lookups

dy_import_module_symbols("geoip_client.r2py") # used for `show location`

dy_import_module_symbols("serialize.r2py") # used for loadstate and savestate

# Import nmclient.r2py after overloading openconnection.
dy_import_module_symbols("nmclient.r2py")

openconnection = old_openconnection


def update_time():
  """
  <Purpose>
    Updates the time that is used internally for nodemanager communications.

  <Arguments>
    None

  <Side Effects>
    Updates the time within time.r2py.
    If there is no network connection, the local clock is used.

  <Exceptions>
    None

  <Returns>
    None
  """
  # Since we import time.r2py, we will only use the local clock if none of
  # the default time servers respond.

  # LAW: Disabling for now as it seems to be being triggered even though
  # TCP/NTP time updates properly.  Its bringing up false negatives on
  # the unit tests.
  # time_register_method('local', local_updatetime)
  time_updatetime(34612)



def local_updatetime(port):
  """
  <Purpose>
    Callback for time_interface.r2py to update the time that is used
    internally for nodemanager communications.

  <Arguments>
    port:
        The port to update on.  This is not used however.  It is only
        specified to adhere to the function signature expected by
        time_interface.r2py.

  <Side Effects>
    If we reach this function, then it means that other time server updates
    failed.  We will notify the user of the failure, and set time.r2py to
    use the local clock.

  <Exceptions>
    None

  <Returns>
    None
  """
  print 'Time update failed, could not connect to any time servers...'
  print 'Your network connection may be down.'
  print "Falling back to using your computer's local clock."
  print

  # time.time() gives us the # of seconds since 1970, whereas the NTP
  # services gives us the # of seconds since 1900.
  time_settime(time.time() + time_seconds_from_1900_to_1970)



# Saves the current state to file. Helper method for the savestate
# command. (Added by Danny Y. Huang)
def savestate(statefn, handleinfo, host, port, expnum, filename, cmdargs,
              defaulttarget, defaultkeyname, autosave, currentkeyname):

  # obtain the handle info dictionary
  for longname in seash_global_variables.vesselinfo.keys():
    vessel_handle = seash_global_variables.vesselinfo[longname]['handle']
    handleinfo[longname] = nmclient_get_handle_info(vessel_handle)


  state = {}
  state['targets'] = seash_global_variables.targets
  state['keys'] = seash_global_variables.keys
  state['vesselinfo'] = seash_global_variables.vesselinfo
  state['nextid'] = seash_global_variables.nextid
  state['handleinfo'] = handleinfo
  state['host'] = host
  state['port'] = port
  state['expnum'] = expnum
  state['filename'] = filename
  state['cmdargs'] = cmdargs
  state['defaulttarget'] = defaulttarget
  state['defaultkeyname'] = defaultkeyname
  state['autosave'] = autosave
  state['globalseashtimeout'] = seash_global_variables.globalseashtimeout
  state['globaluploadrate'] = seash_global_variables.globaluploadrate


  # serialize states and encrypt
  if seash_global_variables.keys.has_key(defaultkeyname):
    cypher = rsa_encrypt(serialize_serializedata(state), seash_global_variables.keys[currentkeyname]['publickey'])
  else:
    raise seash_exceptions.UserError("The keyname '" + defaultkeyname + "' is not loaded.")


  # writing encrypted serialized states to file
  # Exceptions are caught outside of the method
  try:
    state_obj = open(statefn, 'w')
    state_obj.write(cypher)
  finally:
    state_obj.close()



def is_immutable_targetname(targetname):
  if targetname.startswith('%') or ':' in targetname:
    return True
  return False


def valid_targetname(targetname):
  if targetname.startswith('%') or ':' in targetname or ' ' in targetname:
    return False
  return True


def fit_string(stringdata, length):
  if len(stringdata) > length:
    return stringdata[:length-3]+'...'
  return stringdata


nextidlock = createlock()
def atomically_get_nextid():

  # mutex around getting an id
  nextidlock.acquire(True)

  myid = seash_global_variables.nextid
  seash_global_variables.nextid = seash_global_variables.nextid + 1

  nextidlock.release()

  return myid



# adds a vessel and returns the new ID...
def add_vessel(longname, keyname, vesselhandle):

  seash_global_variables.vesselinfo[longname] = {}
  seash_global_variables.vesselinfo[longname]['handle'] = vesselhandle
  seash_global_variables.vesselinfo[longname]['keyname'] = keyname
  seash_global_variables.vesselinfo[longname]['IP'] = longname.split(':')[0]
  seash_global_variables.vesselinfo[longname]['port'] = int(longname.split(':')[1])
  seash_global_variables.vesselinfo[longname]['vesselname'] = longname.split(':')[2]
  # miscelaneous information about the vessel (version, nodeID, etc.)
  seash_global_variables.vesselinfo[longname]['information'] = {}

  # set up a reference to myself...
  seash_global_variables.targets[longname] = [longname]

  myid = atomically_get_nextid()

  # add my id...
  seash_global_variables.targets['%'+str(myid)] = [longname]
  seash_global_variables.vesselinfo[longname]['ID'] = '%'+str(myid)

  # add me to %all...
  seash_global_variables.targets['%all'].append(longname)

  return myid




def copy_vessel(longname, newvesselname):

  newhandle = nmclient_duplicatehandle(seash_global_variables.vesselinfo[longname]['handle'])
  newlongname = seash_global_variables.vesselinfo[longname]['IP']+":"+str(seash_global_variables.vesselinfo[longname]['port'])+":"+newvesselname
  add_vessel(newlongname,seash_global_variables.vesselinfo[longname]['keyname'],newhandle)
  return newlongname




def delete_vessel(longname):

  # remove the item...
  del seash_global_variables.vesselinfo[longname]

  # remove the targets that reference it...
  for target in seash_global_variables.targets.copy():

    # if in my list...
    if longname in seash_global_variables.targets[target]:

      # if this is the %num entry or longname entry...
      if ('%' in target and target != '%all') or longname == target:
        del seash_global_variables.targets[target]
        continue

      # otherwise remove the item from the list...
      seash_global_variables.targets[target].remove(longname)




def longnamelist_to_nodelist(longnamelist):

  retlist = []
  for longname in longnamelist:
    nodename = seash_global_variables.vesselinfo[longname]['IP']+":"+str(seash_global_variables.vesselinfo[longname]['port'])
    retlist.append(nodename)

  return retlist




def find_handle_for_node(nodename):

  for longname in seash_global_variables.vesselinfo:
    if longname.rsplit(':',1)[0] == nodename:
      return seash_global_variables.vesselinfo[longname]['handle']

  raise IndexError("Cannot find handle for '"+nodename+"'")





#################### functions that operate on a target

MAX_CONTACT_WORKER_THREAD_COUNT = 10


# This function abstracts out contacting different nodes.   It spawns off
# multiple worker threads to handle the clients...
# by a threaded model in the future...
# NOTE: entries in targetlist are assumed by me to be unique
def contact_targets(targetlist, func,*args):

  phandle = parallelize_initfunction(targetlist, func, MAX_CONTACT_WORKER_THREAD_COUNT, *args)

  while not parallelize_isfunctionfinished(phandle):
    sleep(.1)

  # I'm going to change the format slightly...
  resultdict = parallelize_getresults(phandle)

  # There really shouldn't be any exceptions in any of the routines...
  if resultdict['exception']:
    print "WARNING: ",resultdict['exception']

  # I'm going to convert the format to be targetname (as the key) and
  # a value with the return value...
  retdict = {}
  for nameandretval in resultdict['returned']:
    retdict[nameandretval[0]] = nameandretval[1]

  return retdict




# This function abstracts out contacting different nodes.   It is obsoleted by
# the threaded model...   This code is retained for testing reasons only
def simple_contact_targets(targetlist, func,*args):

  retdict = {}

  # do the function on each target, returning a dict of results.
  for target in targetlist:
    retdict[target] = func(target,*args)

  return retdict




# used in show files
def showfiles_target(longname):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    filedata = nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'],"ListFilesInVessel",vesselname)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True, filedata)








# used in show log
def showlog_target(longname):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    logdata = nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'],"ReadVesselLog",vesselname)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True, logdata)





# used in show resources
def showresources_target(longname):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    resourcedata = nmclient_rawsay(seash_global_variables.vesselinfo[longname]['handle'],"GetVesselResources",vesselname)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True, resourcedata)


# used in show offcut
def showoffcut_target(nodename):

  vesselhandle = find_handle_for_node(nodename)

  try:
    offcutdata = nmclient_rawsay(vesselhandle,"GetOffcutResources")

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True, offcutdata)





def browse_target(node, currentkeyname, targetgroup='browsegood'):

  # NOTE: I almost think I should skip those nodes that I know about from
  # previous browse commands.   Perhaps I should have an option on the browse
  # command?

  host, portstring = node.split(':')
  port = int(portstring)

  # get information about the node's vessels
  try:
    nodehandle = nmclient_createhandle(host, port,
                                       privatekey = seash_global_variables.keys[currentkeyname]['privatekey'],
                                       publickey = seash_global_variables.keys[currentkeyname]['publickey'],
                                       timeout=seash_global_variables.globalseashtimeout)

  except NMClientException,e:
    return (False, str(e))

  try:
    # need to contact the node to get the list of vessels we can perform
    # actions on...
    ownervessels, uservessels = nmclient_listaccessiblevessels(nodehandle,seash_global_variables.keys[currentkeyname]['publickey'])

    retlist = []

    # we should add anything we can access (whether a user or owner vessel)
    for vesselname in ownervessels + uservessels:
      longname = host+":"+str(port)+":"+vesselname

      # if we haven't discovered the vessel previously...
      if longname not in seash_global_variables.targets:
        # set the vesselname in the handle
        newhandle = nmclient_duplicatehandle(nodehandle)
        handleinfo = nmclient_get_handle_info(newhandle)
        handleinfo['vesselname'] = vesselname
        nmclient_set_handle_info(newhandle, handleinfo)

        # then add the vessel to the target list, etc.
        # add_vessel has no race conditions as long as longname is unique
        # (and it should be unique)
        id = add_vessel(longname,currentkeyname,newhandle)
        seash_global_variables.targets[targetgroup].append(longname)

        # and append some information to be printed...
        retlist.append('%'+str(id)+"("+longname+")")



  finally:
    nmclient_destroyhandle(nodehandle)

  return (True, retlist)


def list_or_update_target(longname):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    vesseldict = nmclient_getvesseldict(seash_global_variables.vesselinfo[longname]['handle'])

  except NMClientException, e:
    return (False, str(e))

  else:
    # updates the dictionary of our node information (dictionary used in show,
    # etc.)
    for key in vesseldict['vessels'][vesselname]:
      seash_global_variables.vesselinfo[longname][key] = vesseldict['vessels'][vesselname][key]

    # Update the "information" (version number, etc.)
    del vesseldict['vessels']
    seash_global_variables.vesselinfo[longname]['information'] = vesseldict

    return (True,)




def upload_target(longname, remotefn, filedata):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # add the file data...
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "AddFileToVessel", vesselname, remotefn, filedata)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




def download_target(longname,localfn,remotefn):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # get the file data...
    retrieveddata = nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "RetrieveFileFromVessel", vesselname, remotefn)

  except NMClientException, e:
    return (False, str(e))

  else:
    writefn = localfn+"."+longname.replace(':','_')
    # write to the local filename (replacing ':' with '_')...
    fileobj = open(writefn,"w")
    fileobj.write(retrieveddata)
    fileobj.close()
    # for output...
    return (True, writefn)



def cat_target(longname,remotefn):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # get the file data...
    retrieveddata = nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "RetrieveFileFromVessel", vesselname, remotefn)

  except NMClientException, e:
    return (False, str(e))

  else:
    # and return it..
    return (True, retrieveddata)



def delete_target(longname,remotefn):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # delete the file...
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "DeleteFileInVessel", vesselname, remotefn)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




def start_target(longname, argstring, prog_platform):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  # start the program
  try:
    # Backwards compatibility with old nodemanagers that don't support
    # StartVesselEX
    if prog_platform == "repyV1":
      nmclient_signedsay(
        seash_global_variables.vesselinfo[longname]['handle'],
        "StartVessel", vesselname, argstring)
    else:
      nmclient_signedsay(
        seash_global_variables.vesselinfo[longname]['handle'],
        "StartVesselEx", vesselname, prog_platform, argstring)

  except NMClientException, e:
    print str(e)
    return (False, str(e))

  else:
    return (True,)




def stop_target(longname):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # stop the programs
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "StopVessel", vesselname)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




def reset_target(longname):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # reset the target
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "ResetVessel", vesselname)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




def run_target(longname,filename,filedata, argstring, prog_platform):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "AddFileToVessel", vesselname, filename, filedata)
    # Backwards compatibility with old nodemanagers that don't support
    # StartVesselEX
    if prog_platform == "repyV1":
      nmclient_signedsay(
        seash_global_variables.vesselinfo[longname]['handle'],
        "StartVessel", vesselname, argstring)
    else:
      nmclient_signedsay(
        seash_global_variables.vesselinfo[longname]['handle'],
        "StartVesselEx", vesselname, prog_platform, argstring)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)



# didn't test...
def split_target(longname, resourcedata):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    newvesselnames = nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "SplitVessel", vesselname, resourcedata)

  except NMClientException, e:
    return (False, str(e))

  else:
    newname1 = copy_vessel(longname, newvesselnames.split()[0])
    newname2 = copy_vessel(longname, newvesselnames.split()[1])
    delete_vessel(longname)
    return (True,(newname1,newname2))




# didn't test...
def join_target(nodename,nodedict):

  if len(nodedict[nodename]) < 2:
    # not enough vessels, nothing to do
    return (False, None)


  # I'll iterate through the vessels, joining one with the current
  # (current starts as the first vessel and becomes the "new vessel")
  currentvesselname = seash_global_variables.vesselinfo[nodedict[nodename][0]]['vesselname']
  currentlongname = nodedict[nodename][0]

  # keep a list of what I replace...
  subsumedlist = [currentlongname]

  for longname in nodedict[nodename][1:]:
    vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

    try:
      newvesselname = nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "JoinVessels", currentvesselname, vesselname)

    except NMClientException, e:
      return (False, str(e))

    else:
      newname = copy_vessel(longname, newvesselname)
      delete_vessel(longname)
      delete_vessel(currentlongname)
      subsumedlist.append(longname)
      currentlongname = newname
      currentvesselname = newvesselname


  else:
    return (True, (currentlongname,subsumedlist))




# didn't test...
def setowner_target(longname,newowner):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "ChangeOwner", vesselname, rsa_publickey_to_string(seash_global_variables.keys[newowner]['publickey']))

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




# didn't test...
def setadvertise_target(longname,newadvert):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # do the actual advertisement changes
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "ChangeAdvertise", vesselname, newadvert)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




def setownerinformation_target(longname,newownerinformation):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    # do the actual advertisement changes
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "ChangeOwnerInformation", vesselname, newownerinformation)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




def setusers_target(longname,userkeystring):

  vesselname = seash_global_variables.vesselinfo[longname]['vesselname']

  try:
    nmclient_signedsay(seash_global_variables.vesselinfo[longname]['handle'], "ChangeUsers", vesselname, userkeystring)

  except NMClientException, e:
    return (False, str(e))

  else:
    return (True,)




# Checks if both keys are setup
def check_key_set(name):

  if (name in seash_global_variables.keys and 'publickey' in seash_global_variables.keys[name] and 'privatekey' in seash_global_variables.keys[name] and seash_global_variables.keys[name]['publickey'] and seash_global_variables.keys[name]['privatekey']):

    if not check_key_pair_compatibility(name):
      raise seash_exceptions.UserError("Error: Mis-matched Public/Private key pair!")




# Check the keys to make sure they are compatible, for the given name
def check_key_pair_compatibility(name):

    # Check for both sets of keys
    setPublic = seash_global_variables.keys[name]['publickey']
    setPrivate = seash_global_variables.keys[name]['privatekey']

    # Check for a mis-match
    match = rsa_matching_keys(setPrivate, setPublic)

    return match


# Reload the handles of a node. Used when "loadstate" is invoked. Returns a
# tuple (success, e), where success is a boolean and e is a string of error
# messages. Added by Danny Y. Huang.
def reload_target(longname, handleinfo):

  host, portstring, vesselname = longname.split(':')
  port = int(portstring)

  try:
    priKey = handleinfo[longname]['privatekey']
    pubKey = handleinfo[longname]['publickey']

  except KeyError:
    error = ("Vessel is absent in the handleinfo dictionary.")
    return (False, error)

  # find the user who has these keys
  thiskeyname = ""

  for keyname in seash_global_variables.keys.keys():
    if (seash_global_variables.keys[keyname]['publickey'] == pubKey and
        seash_global_variables.keys[keyname]['privatekey'] == priKey):
      thiskeyname = keyname
      break

  if not thiskeyname:
    return (False, "User with keyname '" + keyname + "' is not found.")

  # create new handle for the vessel
  try:
    vessel_handle = nmclient_createhandle(host, port, privatekey = priKey, publickey = pubKey, timeout=seash_global_variables.globalseashtimeout)

  except NMClientException, error:
    return (False, str(error))


  try:
    nmclient_set_handle_info(vessel_handle, handleinfo[longname])
    seash_global_variables.vesselinfo[longname]['handle'] = vessel_handle

    # hello test to see if the vessel is available
    (ownervessels, uservessels) = nmclient_listaccessiblevessels(vessel_handle, pubKey)
    if not (ownervessels + uservessels):
      return (False, "Vessel is not available for keyname " + keyname + ".")

  except Exception, error:
    # Catching unexpected exceptions
    return (False, "General exception: " + str(error) + ".")

  return (True, "")




# Determines if there's a need to temporarily change the vessel timeout
# to avoid timing out on bad connection speeds when uploading file.
def set_upload_timeout(filedata):

  filesize = len(filedata)
  est_upload_time = filesize / seash_global_variables.globaluploadrate

  # sets the new timeout if necessary
  if est_upload_time > seash_global_variables.globalseashtimeout:

    for longname in seash_global_variables.vesselinfo:
      thisvesselhandle = seash_global_variables.vesselinfo[longname]['handle']
      thisvesselhandledict = nmclient_get_handle_info(thisvesselhandle)
      thisvesselhandledict['timeout'] = est_upload_time
      nmclient_set_handle_info(thisvesselhandle,thisvesselhandledict)




# Resets each vessel's timeout to the value of globalseashtimeout
def reset_vessel_timeout():

  # resets each vessel's timeout to the original values before file upload
  for longname in seash_global_variables.vesselinfo:
    thisvesselhandle = seash_global_variables.vesselinfo[longname]['handle']
    thisvesselhandledict = nmclient_get_handle_info(thisvesselhandle)
    thisvesselhandledict['timeout'] = seash_global_variables.globalseashtimeout
    nmclient_set_handle_info(thisvesselhandle,thisvesselhandledict)


def get_execution_platform(command, filename):
  """
  <Purpose>
    Returns the execution platform based on a best-guess approach using
    the specified command, as well as the a file's extension.  The
    command takes precedence over the file extension.  If the extension
    is not recognized, then it will be assumed that it is repyV2.

  <Arguments>
    command: The command that should be parsed.
    filename: The file whose repy version should be returned.

  <Side Effects>
    None

  <Exceptions>
    None

  <Returns>
    A string indicating which version of repy a program is in, based on
    its file extension.  This will be either "v1" or "v2".
  """
  if command.endswith('v2'):
    return 'repyV2'
  elif command.endswith('v1'):
    return 'repyV1'

  # Information on extensions for repy programs can be found on #1286.
  if filename.endswith('.r2py'):
    return 'repyV2'
  else:
    return 'repyV1'


def print_vessel_errors(retdict):
  """
  <Purpose>
    Prints out any errors that occurred while performing an action on vessels,
    in a human readable way.

    Errors will be printed out in the following format:
    description [reason]
    Affected vessels: nodelist

    To define a new error, add the following entry to ERROR_RESPONSES in this
    function:
      'error_identifier': {
        'error': 'description for the error',
        'reason': 'reason for the error' (optional).
    'error_identifier'
      This is the substring of the error that can be used to identify it.
      Longer identifiers will have a higher priority over shorter identifiers.
      For example, authentication errors could be identified using the string
      'Insufficient Permissions'.
    'error'
      This is where you put the description for the error to show to the user.
    'reason' (optional)
      This is where you put clarification for the error to ease the user.
      Additionally, you may put things that they could do to fix the error here,
      if applicable. If you don't want to show a reason, don't include this key
      in the dictionary.
      Examples when you would not put a reason is if you received a timeout,
      since the user can't do anything to fix them.

  <Arguments>
    retdict:
      A list of longnames mapped against a tuple (Success?, Message/Errortext).

  <Side Effects>
    Prints error messages onto the screen. See documentation for ERROR_RESPONSES
    for more information.

  <Exceptions>
    Exception

  <Return>
    None

  """
  ERROR_RESPONSES = {
    "Node Manager error 'Insufficient Permissions'": {
      'error': "You lack sufficient permissions to perform this action.",
      'reason': "Did you release the resource(s) by accident?"},
    'timed out': {
      'error':'Connection timed out.'},
    "file not found": {
      'error': "The specified file(s) could not be found.",
      'reason': "Please check the filename."},
    "Node Manager error 'Programming language platform is not supported.'": {
      'error': "Requested platform is not supported by the target vessel."},
    }

  # A dictionary mapping error identifiers to a list of vessels that share
  # that error.
  error_longnames = {}

  for longname in retdict:
    # if the first item is true, then there is no error.
    if not retdict[longname][0]:
      matches = []
      # Loop until we find the response
      for error_string in ERROR_RESPONSES:
        if error_string.lower() in retdict[longname][1].lower():
          # This is the first match
          if not matches:
            matches = [error_string]
          else:
            # This is a better match, forget about the previous matches
            if len(error_string) > len(matches[0]):
              matches = [error_string]

            elif len(error_string) == len(matches[0]):
              matches.append(error_string)

      # If there isn't a match, use the error string as an error identifier
      if not matches:
        errorid = retdict[longname][1]

      else:
        # There should not be more than 1 match for any error.
        # If there is, log the error to a file.
        if len(matches) != 1:
          errfile = open('seasherrors.txt', 'a')
          errorstring = "Multiple matches with same priority:" + '\n'.join(matches)
          errfile.write(errorstring)
          errfile.close()
          raise Exception(errorstring)
        errorid = matches[0]

      # Create the longname list if it doesn't already exist
      if errorid not in error_longnames:
        error_longnames[errorid] = []
      error_longnames[errorid].append(longname)


  # Print the errors out
  for errorid in error_longnames:
    # Represent the list of nodes in a human readable way.
    nodestring = ''
    for node in error_longnames[errorid]:
      # This is the first node
      if node == error_longnames[errorid][0]:
        divider = ''
      # This is a node in the middle
      elif node != error_longnames[errorid][-1]:
        divider = ', '
      # This is the last node
      else:
        # We will always have at least 2 nodes at this point, since if there
        # is only one node, it will be treated as the first node. Therefore,
        # we only have two cases, where there are exactly 2 nodes, or more than
        # 2 nodes.
        # If we have two nodes, we want: "node_1 and node_2".
        # Otherwise, we want: "node_1, node_2, ..., and node_n".
        divider = " and "
        if len(error_longnames[errorid]) > 2:
          divider = ',' + divider
      nodestring += divider + node

    if errorid in ERROR_RESPONSES:
      print ERROR_RESPONSES[errorid]['error'],
      if 'reason' in ERROR_RESPONSES[errorid]:
        print ERROR_RESPONSES[errorid]['reason']
      else:
        # Caret is still on the same line as the list of nodes
        print
    else:
      # Unknown error.
      print "An error occurred: " + errorid
    print "Affected vessels:", nodestring + '.'


def get_commands_from_commanddict(commanddict):
  """
  <Purpose>
    Extracts the commands that are contained in the command dictionary.
    The arguments of these commands are not included.

  <Arguments>
    commanddict: A command dictionary in the format specified in
      seash_dictionary.

  <Exceptions>
    None

  <Side Effects>
    None

  <Return>
    A list of commands that are in the commanddict.
  """
  if not commanddict:
    return []

  commands = []
  for command in commanddict:
    has_user_argument = False
    subcommands = get_commands_from_commanddict(commanddict[command]['children'])
    if subcommands:
      for subcommand in subcommands:
        commands.append(command + " " + subcommand)
    else:
      commands.append(command)
  return commands


def remove_files(files):
  # We can't rely on querying the OS for the existence of the files, as
  # someone could delete the files after we check and before we delete.
  for fn in files:
    try:
      os.remove(fn)
    except OSError, e:
      # Everything but 'file not found' is a serious error here!
      if e.errno != errno.ENOENT:
        raise



"""
Author: Alan Loh
Module: Library of command callback methods. A large portion of the code were
        directly transferred here from the original seash. Each command function 
        is assigned to the appropriate levels of the respective command's 
        dictionary, and are called upon by command_parser.py's dispatch method 
        when executing a user's command input.

All callback methods should take in two arguments: 
-a command dictionary, input_dict, that holds and reflects the user's command input
-the environment_dict holding the following variables:
  host
  port
  expnum 
  filename 
  cmdargs 
  defaulttarget 
  defaultkeyname
  currenttarget
  currentkeyname
  handleinfo
  autosave

In order to pull user-inputed arguments from the command dictionary, the 
following code is used throughout for consistency and ease:

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename input
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

...and command_key will either contain the user's file name, found by searching 
for the appropriate command dictionary with the name 'filename' and taking the 
key associated with it, or it will hold the key of the last command dictionary 
in the input dictionary chain. It is important that the name of the argument 
command dictionary is unique and spelled correctly.

Method names should reflect the expected command string used to executed that 
command dictionary's command callback. For example, the command "show resources
to file"'s method should be called show_resources_to_filename.

Commands with slight variations in function depending on how the user inputted 
it should use a separate method for each variation in its command string. For 
example, 'browse' would have the method named 'browse', while 'browse [type]' 
would have a separate method called 'browse_args'
"""

# For access to additional methods in command handling
import seash_helper
# For access to various global variables kept track of throughout a seash session
import seash_global_variables
# To be able to modify the command dictionary
import seash_dictionary
# To be able to look up what a module commanddict contains (for modulehelp)
import seash_modules

# To be able to throw certain exceptions designed specifically for seash
import seash_exceptions
import seash_modules

import os.path
import sys

from repyportability import *
add_dy_support(locals())

# dy_import_module_symbols("nmclient.r2py")

# This is a temporary workaround for the problem where nmclient is being
# initialized multiple times by dylink.r2py when it is imported multiple
# times.  The handles returned by the nmclient imported here will be
# completely different from the set of handles that the nmclient
# imported in seash_helper knows about.  This needs to be removed once
# dylink is properly fixed. See #1319 for more information.

nmclient_createhandle = seash_helper.nmclient_createhandle
nmclient_listaccessiblevessels = seash_helper.nmclient_listaccessiblevessels
nmclient_duplicatehandle = seash_helper.nmclient_duplicatehandle
nmclient_get_handle_info = seash_helper.nmclient_get_handle_info
nmclient_set_handle_info = seash_helper.nmclient_set_handle_info

dy_import_module_symbols("time.r2py")

dy_import_module_symbols("rsa.r2py")

dy_import_module_symbols("listops.r2py")

dy_import_module_symbols("parallelize.r2py")

dy_import_module_symbols("domainnameinfo.r2py")

dy_import_module_symbols("advertise.r2py")   #  used to do OpenDHT lookups

dy_import_module_symbols("serialize.r2py") # used for loadstate and savestate

# The versions of Seattle that we officially support.
SUPPORTED_PROG_PLATFORMS = ["repyV1", "repyV2"]


# set the target, then handle other operations
def on_target(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's target name
  while input_dict[command_key]['name'] is not 'ontarget':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  if command_key not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error: Unknown target '"+command_key+"'")

  # set the target and strip the rest...
  environment_dict['currenttarget'] = command_key

  # Set default if there's no follow-up commands
  if not input_dict[command_key]['children']:
    environment_dict['defaulttarget'] = environment_dict['currenttarget']




# set the keys, then handle other operations
def as_keyname(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's keyname input
  while input_dict[command_key]['name'] is not 'askeyname':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]


  if command_key not in seash_global_variables.keys:
    raise seash_exceptions.UserError("Error: Unknown identity '"+command_key+"'")

            
  # set the target and strip the rest...
  environment_dict['currentkeyname'] = command_key

  # Set default if there's no follow-up commands
  if not input_dict[command_key]['children']:
    environment_dict['defaultkeyname'] = environment_dict['currentkeyname']



###Show Commands###

# show            -- Does nothing
def show(input_dict, environment_dict):
  pass




# show info       -- Display general information about the vessels
def show_info(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")


  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:

    if seash_global_variables.vesselinfo[longname]['information']:
      print longname, seash_global_variables.vesselinfo[longname]['information']

    else:
      print longname, "has no information (try 'update' or 'list')"




# show targets    -- Display a list of targets
def show_targets(input_dict, environment_dict):

  for target in seash_global_variables.targets:

    if len(seash_global_variables.targets[target]) == 0:
      print target, "(empty)"
      continue

    # this is a vesselentry
    if target == seash_global_variables.targets[target][0]:
      continue

    print target, seash_global_variables.targets[target]



# show groups    -- Display a list of groups 
def show_groups(input_dict, environment_dict):

  for target in seash_global_variables.targets:

    if target == "%all":
      print target, seash_global_variables.targets[target]
      continue
	
    # this is individual target number
    if target.startswith('%'):
      continue
	
    if len(seash_global_variables.targets[target]) == 0:
      print target, seash_global_variables.targets[target]
      continue
	  
    # this is a vesselentry
    if target == seash_global_variables.targets[target][0]:
      continue
	  
    print target, seash_global_variables.targets[target]



# show keys       -- Display the known keys
def show_keys(input_dict, environment_dict):

  for identity in seash_global_variables.keys:

    print identity,seash_global_variables.keys[identity]['publickey'],seash_global_variables.keys[identity]['privatekey']




# show identities -- Display the known identities
def show_identities(input_dict, environment_dict):

  for keyname in seash_global_variables.keys:

    print keyname,

    if seash_global_variables.keys[keyname]['publickey']:
      print "PUB",

    if seash_global_variables.keys[keyname]['privatekey']:
      print "PRIV",

    print




# show users      -- Display the user keys for the vessels
def show_users(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:

    if 'userkeys' in seash_global_variables.vesselinfo[longname]:
      if seash_global_variables.vesselinfo[longname]['userkeys'] == []:
        print longname,"(no keys)"
        continue

      print longname,

      # we'd like to say 'joe's public key' instead of '3453 2323...'
      for key in seash_global_variables.vesselinfo[longname]['userkeys']:

        for identity in seash_global_variables.keys:

          if seash_global_variables.keys[identity]['publickey'] == key:
            print identity,
            break

        else:
          print seash_helper.fit_string(rsa_publickey_to_string(key),15),

      print

    else:
      print longname, "has no information (try 'update' or 'list')"
    



# show ownerinfo  -- Display owner information for the vessels
def show_ownerinfo(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:

    if 'ownerinfo' in seash_global_variables.vesselinfo[longname]:
      print longname, "'"+seash_global_variables.vesselinfo[longname]['ownerinfo']+"'"
      # list all of the info...

    else:
      print longname, "has no information (try 'update' or 'list')"




# show advertise  -- Display advertisement information about the vessels
def show_advertise(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:

    if 'advertise' in seash_global_variables.vesselinfo[longname]:

      if seash_global_variables.vesselinfo[longname]['advertise']:
        print longname, "on"

      else:
        print longname, "off"

    # list all of the info...
    else:
      print longname, "has no information (try 'update' or 'list')"




# show owner      -- Display a vessel's owner
def show_owner(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:

    if 'ownerkey' in seash_global_variables.vesselinfo[longname]:
      # we'd like to say 'joe public key' instead of '3453 2323...'
      ownerkey = seash_global_variables.vesselinfo[longname]['ownerkey']

      for identity in seash_global_variables.keys:

        if seash_global_variables.keys[identity]['publickey'] == ownerkey:
          print longname, identity+" pubkey"
          break

      else:
        print longname, seash_helper.fit_string(rsa_publickey_to_string(ownerkey),15)

    else:
      print longname, "has no information (try 'update' or 'list')"




# show files      -- Display a list of files in the vessel (*)
def show_files(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  # print the list of files in the vessels (seash method)
  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']], seash_helper.showfiles_target)

  goodlist = []
  faillist = []

  for longname in retdict:

    # True means it worked
    if retdict[longname][0]:
      # let's sort the files...
      fileliststring = retdict[longname][1]
      filelist = fileliststring.split()
      filelist.sort()
      fileliststring = " ".join(filelist)
      print "Files on '"+longname+"': '"+fileliststring+"'"
      goodlist.append(longname)

    else:
      faillist.append(longname)
  
  seash_helper.print_vessel_errors(retdict)

  if goodlist and faillist:
    seash_global_variables.targets['filesgood'] = goodlist
    seash_global_variables.targets['filesfail'] = faillist
    print "Added group 'filesgood' with "+str(len(seash_global_variables.targets['filesgood']))+" targets and 'filesfail' with "+str(len(seash_global_variables.targets['filesfail']))+" targets"




# show log [to file] -- Display the log from the vessel (*)
def show_log(input_dict, environment_dict):

  writeoutputtofile = False


  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  # print the vessel logs...
  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']], seash_helper.showlog_target)

  goodlist = []
  faillist = []

  for longname in retdict:
    # True means it worked
    if retdict[longname][0]:
      if writeoutputtofile:
        # write to a file if requested.
        outputfilename = outputfileprefix +'.'+ longname
        outputfo = file(outputfilename,'w')
        outputfo.write(retdict[longname][1])
        outputfo.close()
        print "Wrote log as '"+outputfilename+"'."

      else:
        # else print it to the terminal
        print "Log from '"+longname+"':"
        print retdict[longname][1]

      # regardless, this is a good node
      goodlist.append(longname)

    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if goodlist and faillist:
    seash_global_variables.targets['loggood'] = goodlist
    seash_global_variables.targets['logfail'] = faillist
    print "Added group 'loggood' with "+str(len(seash_global_variables.targets['loggood']))+" targets and 'logfail' with "+str(len(seash_global_variables.targets['logfail']))+" targets"




# show log to file
def show_log_to_file(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  writeoutputtofile = True
  # handle '~'
  fileandpath = os.path.expanduser(command_key)
  outputfileprefix = fileandpath


  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  # print the vessel logs...
  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']], seash_helper.showlog_target)

  goodlist = []
  faillist = []

  for longname in retdict:
    # True means it worked
    if retdict[longname][0]:
      if writeoutputtofile:
        # write to a file if requested.
        outputfilename = outputfileprefix +'.'+ longname
        outputfo = file(outputfilename,'w')
        outputfo.write(retdict[longname][1])
        outputfo.close()
        print "Wrote log as '"+outputfilename+"'."

      else:
        # else print it to the terminal
        print "Log from '"+longname+"':"
        print retdict[longname][1]

      # regardless, this is a good node
      goodlist.append(longname)

    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if goodlist and faillist:
    seash_global_variables.targets['loggood'] = goodlist
    seash_global_variables.targets['logfail'] = faillist
    print "Added group 'loggood' with "+str(len(seash_global_variables.targets['loggood']))+" targets and 'logfail' with "+str(len(seash_global_variables.targets['logfail']))+" targets"




# show resources  -- Display the resources / restrictions for the vessel (*)
def show_resources(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']], seash_helper.showresources_target)
  faillist = []
  goodlist = []

  for longname in retdict:

    # True means it worked
    if retdict[longname][0]:
      print "Resource data for '"+longname+"':"
      print retdict[longname][1]
      goodlist.append(longname)

    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if goodlist and faillist:
    seash_global_variables.targets['resourcegood'] = goodlist
    seash_global_variables.targets['resourcefail'] = faillist
    print "Added group 'resourcegood' with "+str(len(seash_global_variables.targets['resourcegood']))+" targets and 'resourcefail' with "+str(len(seash_global_variables.targets['resourcefail']))+" targets"




# show offcut     -- Display the offcut resource for the node (*)
def show_offcut(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")


  # we should only visit a node once...
  nodelist = listops_uniq(seash_helper.longnamelist_to_nodelist(seash_global_variables.targets[environment_dict['currenttarget']]))
  retdict = seash_helper.contact_targets(nodelist, seash_helper.showoffcut_target)

  for nodename in retdict:

    if retdict[nodename][0]:
      print "Offcut resource data for '"+nodename+"':"
      print retdict[nodename][1]

  seash_helper.print_vessel_errors(retdict)




# show ip [to file] -- Display the ip addresses of the nodes
def show_ip(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  # print to the terminal (stdout)
  # write data here...
  outfo = sys.stdout


  # we should only visit a node once...
  printedIPlist = []

  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:
    thisnodeIP = seash_global_variables.vesselinfo[longname]['IP']

    if thisnodeIP not in printedIPlist:
      printedIPlist.append(thisnodeIP)
      print >> outfo, thisnodeIP




# show ip to file
def show_ip_to_file(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  outfo = open(command_key,"w+")


  # we should only visit a node once...
  printedIPlist = []

  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:
    thisnodeIP = seash_global_variables.vesselinfo[longname]['IP']

    if thisnodeIP not in printedIPlist:
      printedIPlist.append(thisnodeIP)
      print >> outfo, thisnodeIP

  # if it's a file, close it...
  outfo.close()




# show hostname        -- Display the hostnames of the nodes
def show_hostname(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Error, command requires a target")


  # we should only visit a node once...
  printedIPlist = []

  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:
    thisnodeIP = seash_global_variables.vesselinfo[longname]['IP']

    if thisnodeIP not in printedIPlist:
      printedIPlist.append(thisnodeIP)
      print thisnodeIP,

      try: 
        nodeinfo = socket.gethostbyaddr(thisnodeIP)

      except (socket.herror,socket.gaierror, socket.timeout, socket.error):
        print 'has unknown host information'

      else:
        print 'is known as',nodeinfo[0]







# show timeout    -- Display the timeout for nodes
def show_timeout(input_dict, environment_dict):
  print seash_global_variables.globalseashtimeout




# show uploadrate    -- Display the file upload rate
def show_uploadrate(input_dict, environment_dict):
  print seash_global_variables.globaluploadrate




# add target (to group)
def add_target(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's target name
  while input_dict[command_key]['name'] is not 'target':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  source = command_key
  dest = environment_dict['currenttarget']


  # okay, now source and dest are set.   Time to add the nodes in source
  # to the group dest...
  if source not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown target '"+source+"'")

  if dest not in seash_global_variables.targets:
    if not seash_helper.valid_targetname(dest):
      raise seash_exceptions.UserError("Error, invalid target name '"+dest+"'")
    seash_global_variables.targets[dest] = []

  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError("Can't modify the contents of '"+dest+"'")

  # source - dest has what we should add (items in source but not dest)
  addlist = listops_difference(seash_global_variables.targets[source],seash_global_variables.targets[dest])

  if len(addlist) == 0:
    raise seash_exceptions.UserError("No targets to add (the target is already in '"+dest+"')")

  for item in addlist:
    seash_global_variables.targets[dest].append(item)




# add target to group
def add_target_to_group(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's target name
  while input_dict[command_key]['name'] is not 'target':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  source = command_key

  # Iterates through the dictionary to retrieve the user's group argument
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  dest = command_key


  # okay, now source and dest are set.   Time to add the nodes in source
  # to the group dest...
  if source not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown target '"+source+"'")

  if dest not in seash_global_variables.targets:
    if not seash_helper.valid_targetname(dest):
      raise seash_exceptions.UserError("Error, invalid target name '"+dest+"'")

    seash_global_variables.targets[dest] = []

  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError("Can't modify the contents of '"+dest+"'")

  # source - dest has what we should add (items in source but not dest)
  addlist = listops_difference(seash_global_variables.targets[source],seash_global_variables.targets[dest])

  if len(addlist) == 0:
    raise seash_exceptions.UserError("No targets to add (the target is already in '"+dest+"')")

  for item in addlist:
    seash_global_variables.targets[dest].append(item)





# add to group
def add_to_group(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's group argument
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  source = environment_dict['currenttarget']
  dest = command_key


  # okay, now source and dest are set.   Time to add the nodes in source
  # to the group dest...
  if source not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown target '"+source+"'")

  if dest not in seash_global_variables.targets:
    if not seash_helper.valid_targetname(dest):
      raise seash_exceptions.UserError("Error, invalid target name '"+dest+"'")
    seash_global_variables.targets[dest] = []

  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError("Can't modify the contents of '"+dest+"'")

  # source - dest has what we should add (items in source but not dest)
  addlist = listops_difference(seash_global_variables.targets[source],seash_global_variables.targets[dest])

  if len(addlist) == 0:
    raise seash_exceptions.UserError("No targets to add (the target is already in '"+dest+"')")

  for item in addlist:
    seash_global_variables.targets[dest].append(item)





# remove target (from group)
def remove_target(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's target name
  while input_dict[command_key]['name'] is not 'target':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  source = command_key
  dest = environment_dict['currenttarget']



  # time to check args and do the ops
  if source not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown target '"+source+"'")

  if dest not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown group '"+dest+"'")

  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError("Can't modify the contents of '"+dest+"'")

  # find the items to remove (the items in both dest and source)
  removelist = listops_intersect(seash_global_variables.targets[dest],seash_global_variables.targets[source])

  if len(removelist) == 0:
    raise seash_exceptions.UserError("No targets to remove (no items from '"+source+"' are in '"+dest+"')")

  # it's okay to end up with an empty group.   We'll leave it...
  for item in removelist:
    seash_global_variables.targets[dest].remove(item)




# remove target from group
def remove_target_from_group(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's target name
  while input_dict[command_key]['name'] is not 'target':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  source = command_key

  # Iterates through the dictionary to retrieve the user's group name
  while input_dict[command_key]['name'] is not 'group':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  dest = command_key


  # time to check args and do the ops
  if source not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown target '"+source+"'")

  if dest not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown group '"+dest+"'")

  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError("Can't modify the contents of '"+dest+"'")

  # find the items to remove (the items in both dest and source)
  removelist = listops_intersect(seash_global_variables.targets[dest],seash_global_variables.targets[source])

  if len(removelist) == 0:
    raise seash_exceptions.UserError("No targets to remove (no items from '"+source+"' are in '"+dest+"')")

  # it's okay to end up with an empty group.   We'll leave it...
  for item in removelist:
    seash_global_variables.targets[dest].remove(item)




# remove from group
def remove_from_group(input_dict, environment_dict):

  source = environment_dict['currenttarget']

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's group name
  while input_dict[command_key]['name'] is not 'group':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  dest = command_key


  # time to check args and do the ops
  if source not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown target '"+source+"'")

  if dest not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown group '"+dest+"'")

  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError("Can't modify the contents of '"+dest+"'")

  # find the items to remove (the items in both dest and source)
  removelist = listops_intersect(seash_global_variables.targets[dest],seash_global_variables.targets[source])

  if len(removelist) == 0:
    raise seash_exceptions.UserError("No targets to remove (no items from '"+source+"' are in '"+dest+"')")

  # it's okay to end up with an empty group.   We'll leave it...
  for item in removelist:
    seash_global_variables.targets[dest].remove(item)




# move target to group
def move_target_to_group(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's target name
  while input_dict[command_key]['name'] is not 'target':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  moving = command_key
  source = environment_dict['currenttarget']

  # Iterates through the dictionary to retrieve the user's group name
  while input_dict[command_key]['name'] is not 'group':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  dest = command_key


  # check args...
  if source not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown group '"+source+"'")

  if moving not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown group '"+moving+"'")

  if dest not in seash_global_variables.targets:
    raise seash_exceptions.UserError("Error, unknown group '"+dest+"'")


  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError("Can't modify the contents of '"+source+"'")

  if seash_helper.is_immutable_targetname(dest):
    raise seash_exceptions.UserError( "Can't modify the contents of '"+dest+"'")

  removelist = listops_intersect(seash_global_variables.targets[source], seash_global_variables.targets[moving])
  if len(removelist) == 0:
    raise seash_exceptions.UserError("Error, '"+moving+"' is not in '"+source+"'")

  addlist = listops_difference(removelist, seash_global_variables.targets[dest])
  if len(addlist) == 0:
    raise seash_exceptions.UserError("Error, the common items between '"+source+"' and '"+moving+"' are already in '"+dest+"'")

  for item in removelist:
    seash_global_variables.targets[source].remove(item)

  for item in addlist:
    seash_global_variables.targets[dest].append(item)








# contact host:port[:vessel] -- Communicate with a node
def contact(input_dict, environment_dict):

  if environment_dict['currentkeyname'] == None or not seash_global_variables.keys[environment_dict['currentkeyname']]['publickey']:
    raise seash_exceptions.UserError("Error, must contact as an identity")

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  if len(command_key.split(':')) == 2:
    environment_dict['host'], portstring = command_key.split(':')
    environment_dict['port'] = int(portstring)
    vesselname = None

  elif len(command_key.split(':')) == 3:
    environment_dict['host'], portstring,vesselname = command_key.split(':')
    environment_dict['port'] = int(portstring)

  else:
    raise seash_exceptions.UserError("Error, usage is contact host:port[:vessel]")


  # get information about the node's vessels
  thishandle = nmclient_createhandle(environment_dict['host'], environment_dict['port'], privatekey = seash_global_variables.keys[environment_dict['currentkeyname']]['privatekey'], publickey = seash_global_variables.keys[environment_dict['currentkeyname']]['publickey'], vesselid = vesselname, timeout = seash_global_variables.globalseashtimeout)
  ownervessels, uservessels = nmclient_listaccessiblevessels(thishandle,seash_global_variables.keys[environment_dict['currentkeyname']]['publickey'])

  newidlist = []
  # determine if we control the specified vessel...
  if vesselname:
    if vesselname in ownervessels or vesselname in uservessels:
      longname = environment_dict['host']+":"+str(environment_dict['port'])+":"+vesselname
      # no need to set the vesselname, we did so above...
      id = seash_helper.add_vessel(longname,environment_dict['currentkeyname'],thishandle)
      newidlist.append('%'+str(id)+"("+longname+")")

    else:
      raise seash_exceptions.UserError("Error, cannot access vessel '"+vesselname+"'")

  # we should add anything we can access
  else:
    for vesselname in ownervessels:

      longname = environment_dict['host']+":"+str(environment_dict['port'])+":"+vesselname

      if longname not in seash_global_variables.targets:
        # set the vesselname
        # NOTE: we leak handles (no cleanup of thishandle).   
        # I think we don't care...
        newhandle = nmclient_duplicatehandle(thishandle)
        environment_dict['handleinfo'] = nmclient_get_handle_info(newhandle)
        environment_dict['handleinfo']['vesselname'] = vesselname
        nmclient_set_handle_info(newhandle, environment_dict['handleinfo'])

        id = seash_helper.add_vessel(longname,environment_dict['currentkeyname'],newhandle)
        newidlist.append('%'+str(id)+"("+longname+")")


    for vesselname in uservessels:

      longname = environment_dict['host']+":"+str(environment_dict['port'])+":"+vesselname

      if longname not in seash_global_variables.targets:
        # set the vesselname
        # NOTE: we leak handles (no cleanup of thishandle).   
        # I think we don't care...
        newhandle = nmclient_duplicatehandle(thishandle)
        environment_dict['handleinfo'] = nmclient_get_handle_info(newhandle)
        environment_dict['handleinfo']['vesselname'] = vesselname
        nmclient_set_handle_info(newhandle, environment_dict['handleinfo'])

        id = seash_helper.add_vessel(longname,environment_dict['currentkeyname'],newhandle)
        newidlist.append('%'+str(id)+"("+longname+")")


  # tell the user what we did...
  if len(newidlist) == 0:
    print "Could not add any targets."

  else:
    print "Added targets: "+", ".join(newidlist)
            
  





# browse                               -- Find experiments I can control
def browse(input_dict, environment_dict):
 
  if environment_dict['currentkeyname'] == None or not seash_global_variables.keys[environment_dict['currentkeyname']]['publickey']:
    raise seash_exceptions.UserError("Error, must browse as an identity with a public key")


  try:
    nodelist = advertise_lookup(seash_global_variables.keys[environment_dict['currentkeyname']]['publickey'], graceperiod = 3)
  except (AdvertiseError, TimeoutError), e:
    # print the error and return to the user.   Let them decide what to do.
    print "Error:",e
    print "Retry or check network connection."
    return
    


  # If there are no vessels for a user, the lookup may return ''
  for nodename in nodelist[:]:
    if nodename == '':
      nodelist.remove(nodename)

  # we'll output a message about the new keys later...
  newidlist = []

  faillist = []

  seash_global_variables.targets['browsegood'] = []

  print nodelist

  # currently, if I browse more than once, I look up everything again...
  retdict = seash_helper.contact_targets(nodelist,seash_helper.browse_target, environment_dict['currentkeyname'])

  # parse the output so we can print out something intelligible
  for nodename in retdict:

    if retdict[nodename][0]:
      newidlist = newidlist + retdict[nodename][1]

    else:
      faillist.append(nodename)

  seash_helper.print_vessel_errors(retdict)

  if len(newidlist) == 0:
    print "Could not add any new targets."
  else:
    print "Added targets: "+", ".join(newidlist)

  if len(seash_global_variables.targets['browsegood']) > 0:
    print "Added group 'browsegood' with "+str(len(seash_global_variables.targets['browsegood']))+" targets"



# browse [arg]
def browse_arg(input_dict, environment_dict):

  if environment_dict['currentkeyname'] == None or not seash_global_variables.keys[environment_dict['currentkeyname']]['publickey']:
    raise seash_exceptions.UserError("Error, must browse as an identity with a public key")


  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument input
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  type_list = command_key.split()


  try:
    # they are trying to only do some types of lookup...
    nodelist = advertise_lookup(seash_global_variables.keys[environment_dict['currentkeyname']]['publickey'],lookuptype=type_list)
  except (AdvertiseError, TimeoutError), e:
    # print the error and return to the user.   Let them decide what to do.
    print "Error:",e
    print "Retry or check network connection."
    return
    


  # If there are no vessels for a user, the lookup may return ''
  for nodename in nodelist[:]:
    if nodename == '':
      nodelist.remove(nodename)

  # we'll output a message about the new keys later...
  newidlist = []

  faillist = []

  seash_global_variables.targets['browsegood'] = []

  print nodelist
  # currently, if I browse more than once, I look up everything again...
  retdict = seash_helper.contact_targets(nodelist,seash_helper.browse_target, environment_dict['currentkeyname'])

  # parse the output so we can print out something intelligible
  for nodename in retdict:

    if retdict[nodename][0]:
      newidlist = newidlist + retdict[nodename][1]

  seash_helper.print_vessel_errors(retdict)

  if len(newidlist) == 0:
    print "Could not add any new targets."
  else:
    print "Added targets: "+", ".join(newidlist)

  if len(seash_global_variables.targets['browsegood']) > 0:
    print "Added group 'browsegood' with "+str(len(seash_global_variables.targets['browsegood']))+" targets"




# genkeys filename                             -- creates keys
def genkeys_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename input
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  keylength = 1024
  # expand '~'
  fileandpath = os.path.expanduser(command_key)
  keyname = os.path.basename(fileandpath)
  pubkeyfn = fileandpath+'.publickey'
  privkeyfn = fileandpath+'.privatekey'

  # RepyV2's API does not allow us to truncate a file in-place.
  # Therefore, we need to make sure we're creating a new file, otherwise
  # the contents of the public/private key files that we create will be
  # corrupt.

  seash_helper.remove_files([pubkeyfn, privkeyfn])

  # do the actual generation (will take a while)
  newkeys = rsa_gen_pubpriv_keys(keylength)

  rsa_privatekey_to_file(newkeys[1],privkeyfn)
  rsa_publickey_to_file(newkeys[0],pubkeyfn)
  seash_global_variables.keys[keyname] = {'publickey':newkeys[0], 'privatekey':newkeys[1]}

  print "Created identity '"+keyname+"'"




# genkeys filename [len]
def genkeys_filename_len(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename input
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  # expand '~'
  fileandpath = os.path.expanduser(command_key)
  keyname = os.path.basename(fileandpath)
  pubkeyfn = fileandpath+'.publickey'
  privkeyfn = fileandpath+'.privatekey'


  # Iterates through the dictionary to retrieve the user's length argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  keylength = int(command_key)


  # do the actual generation (will take a while)
  newkeys = rsa_gen_pubpriv_keys(keylength)

  # RepyV2's API does not allow us to truncate a file in-place.
  # Therefore, we need to make sure we're creating a new file, otherwise
  # the contents of the public/private key files that we create will be
  # corrupt.

  seash_helper.remove_files([pubkeyfn, privkeyfn])

  rsa_privatekey_to_file(newkeys[1],privkeyfn)
  rsa_publickey_to_file(newkeys[0],pubkeyfn)
  seash_global_variables.keys[keyname] = {'publickey':newkeys[0], 'privatekey':newkeys[1]}

  print "Created identity '"+keyname+"'"




# genkeys filename [len] [as identity]
def genkeys_filename_len_as_identity(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename input
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  pubkeyfn = command_key+'.publickey'
  privkeyfn = command_key+'.privatekey'


  # Iterates through the dictionary to retrieve the user's length argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  keylength = int(command_key)


  # Iterates through the dictionary to retrieve the user's keyname argument
  while input_dict[command_key]['name'] is not 'keyname':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  keyname = command_key


  # do the actual generation (will take a while)
  newkeys = rsa_gen_pubpriv_keys(keylength)

  # RepyV2's API does not allow us to truncate a file in-place.
  # Therefore, we need to make sure we're creating a new file, otherwise
  # the contents of the public/private key files that we create will be
  # corrupt.

  seash_helper.remove_files([pubkeyfn, privkeyfn])

  rsa_privatekey_to_file(newkeys[1],privkeyfn)
  rsa_publickey_to_file(newkeys[0],pubkeyfn)
  seash_global_variables.keys[keyname] = {'publickey':newkeys[0], 'privatekey':newkeys[1]}

  print "Created identity '"+keyname+"'"




# genkeys filename [as identity]
def genkeys_filename_as_identity(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename input
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  keylength = 1024
  pubkeyfn = command_key + '.publickey'
  privkeyfn = command_key + '.privatekey'


  # Iterates through the dictionary to retrieve the user's keyname argument
  while input_dict[command_key]['name'] is not 'keyname':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  keyname = command_key

  # do the actual generation (will take a while)
  newkeys = rsa_gen_pubpriv_keys(keylength)

  # RepyV2's API does not allow us to truncate a file in-place.
  # Therefore, we need to make sure we're creating a new file, otherwise
  # the contents of the public/private key files that we create will be
  # corrupt.

  seash_helper.remove_files([pubkeyfn, privkeyfn])

  rsa_privatekey_to_file(newkeys[1],privkeyfn)
  rsa_publickey_to_file(newkeys[0],pubkeyfn)
  seash_global_variables.keys[keyname] = {'publickey':newkeys[0], 'privatekey':newkeys[1]}

  print "Created identity '"+keyname+"'"
  




# loadpub filename [as identity]                    -- loads a public key
def loadpub_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  # they typed 'loadpub foo.publickey'
  if command_key.endswith('.publickey'):
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    keyname = os.path.basename(fileandpath[:len('.publickey')])
    pubkeyfn = fileandpath
  else:
    # they typed 'loadpub foo'
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    keyname = os.path.basename(fileandpath)
    pubkeyfn = fileandpath+'.publickey'

  # load the key and update the table...
  pubkey = rsa_file_to_publickey(pubkeyfn)

  if keyname not in seash_global_variables.keys:
    seash_global_variables.keys[keyname] = {'publickey':pubkey, 'privatekey':None}
  else:
    seash_global_variables.keys[keyname]['publickey'] = pubkey

    # Check the keys, on error reverse the change and re-raise
    try:
      seash_helper.check_key_set(keyname)

    except:
      seash_global_variables.keys[keyname]['publickey'] = None
      raise




# loadpub keyname as identity
def loadpub_filename_as(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  # they typed 'loadpub foo.publickey'
  if command_key.endswith('.publickey'):
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    pubkeyfn = fileandpath
  else:
    # they typed 'loadpub foo'
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    pubkeyfn = fileandpath+'.publickey'

# Iterates through the dictionary to retrieve the user's keyname argument
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  keyname = command_key

  # load the key and update the table...
  pubkey = rsa_file_to_publickey(pubkeyfn)
  if keyname not in seash_global_variables.keys:
    seash_global_variables.keys[keyname] = {'publickey':pubkey, 'privatekey':None}
  else:
    seash_global_variables.keys[keyname]['publickey'] = pubkey

    # Check the keys, on error reverse the change and re-raise
    try:
      seash_helper.check_key_set(keyname)
    except:
      seash_global_variables.keys[keyname]['publickey'] = None
      raise




# loadpriv filename [as identity]                    -- loads a private key
def loadpriv_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  # they typed 'loadpriv foo.privatekey'
  if command_key.endswith('.privatekey'):
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    privkeyfn = fileandpath+'.privatekey'
    keyname = os.path.basename(fileandpath[:len('.privatekey')])
  else:
    # they typed 'loadpriv foo'
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    privkeyfn = fileandpath+'.privatekey'
    keyname = os.path.basename(fileandpath)


  # load the key and update the table...
  privkey = rsa_file_to_privatekey(privkeyfn)
  if keyname not in seash_global_variables.keys:
    seash_global_variables.keys[keyname] = {'privatekey':privkey, 'publickey':None}
  else:
    seash_global_variables.keys[keyname]['privatekey'] = privkey

    # Check the keys, on error reverse the change and re-raise
    try:
      seash_helper.check_key_set(keyname)
    except:
      seash_global_variables.keys[keyname]['privatekey'] = None
      raise
          



# loadpriv filename as identity
def loadpriv_filename_as(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  # they typed 'loadpriv foo.privatekey'
  if command_key.endswith('.privatekey'):
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    privkeyfn = fileandpath
  else:
    # they typed 'loadpriv foo'
    # handle '~'
    fileandpath = os.path.expanduser(command_key)
    privkeyfn = fileandpath+'.publickey'

  # Iterates through the dictionary to retrieve the user's keyname argument
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  keyname = command_key


  # load the key and update the table...
  privkey = rsa_file_to_privatekey(privkeyfn)
  if keyname not in seash_global_variables.keys:
    seash_global_variables.keys[keyname] = {'privatekey':privkey, 'publickey':None}
  else:
    seash_global_variables.keys[keyname]['privatekey'] = privkey

    # Check the keys, on error reverse the change and re-raise
    try:
      seash_helper.check_key_set(keyname)
    except:
      seash_global_variables.keys[keyname]['privatekey'] = None
      raise




# loadkeys filename [as identity]                    -- loads a private key
def loadkeys_keyname(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  if command_key.endswith('publickey') or command_key.endswith('privatekey'):
    print 'Warning: Trying to load a keypair named "'+command_key+'.publickey" and "'+command_key+'.privatekey"'


  # the user input may have a directory or tilde in it.   The key name 
  # shouldn't have either.
  fileandpath = os.path.expanduser(command_key)
  keyname = os.path.basename(fileandpath)
  privkeyfn = fileandpath+'.privatekey'
  pubkeyfn = fileandpath+'.publickey'




  # load the keys and update the table...
  try:
    privkey = rsa_file_to_privatekey(privkeyfn)
  except (OSError, IOError), e:
    raise seash_exceptions.UserError("Cannot locate private key '"+privkeyfn+"'.\nDetailed error: '"+str(e)+"'.")

  try:
    pubkey = rsa_file_to_publickey(pubkeyfn)
  except (OSError, IOError), e:
    raise seash_exceptions.UserError("Cannot locate private key '"+privkeyfn+"'.\nDetailed error: '"+str(e)+"'.")

  seash_global_variables.keys[keyname] = {'privatekey':privkey, 'publickey':pubkey}


  # Check the keys, on error reverse the change and re-raise
  try:
    seash_helper.check_key_set(keyname)
  except:
    del seash_global_variables.keys[keyname]
    raise




# loadkeys keyname as identity
def loadkeys_keyname_as(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  if command_key.endswith('publickey') or command_key.endswith('privatekey'):
    print 'Warning: Trying to load a keypair named "'+command_key+'.publickey" and "'+command_key+'.privatekey"'


  # the user input may have a directory or tilde in it.   The key name 
  # shouldn't have either.
  fileandpath = os.path.expanduser(command_key)
  privkeyfn = fileandpath+'.privatekey'
  pubkeyfn = fileandpath+'.publickey'
  

  # Iterates through the dictionary to retrieve the user's keyname argument
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  keyname = command_key


  # load the keys and update the table...
  try:
    privkey = rsa_file_to_privatekey(privkeyfn)
  except (OSError, IOError), e:
    raise seash_exceptions.UserError("Cannot locate private key '"+privkeyfn+"'.\nDetailed error: '"+str(e)+"'.")

  try:
    pubkey = rsa_file_to_publickey(pubkeyfn)
  except (OSError, IOError), e:
    raise seash_exceptions.UserError("Cannot locate private key '"+privkeyfn+"'.\nDetailed error: '"+str(e)+"'.")

  seash_global_variables.keys[keyname] = {'privatekey':privkey, 'publickey':pubkey}


  # Check the keys, on error reverse the change and re-raise
  try:
    seash_helper.check_key_set(keyname)
  except:
    del seash_global_variables.keys[keyname]
    raise



"""
 list               -- Update and display information about the vessels

 output looks similar to:
  ID Own                       Name     Status              Owner Information
  %1  *       128.208.3.173:1224:v5      Fresh                               
  %2  *        128.208.3.86:1224:v2      Fresh                               
  %3          234.17.98.23:53322:v5    Stopped               Chord experiment

"""
def list(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  # update information about the vessels...
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.list_or_update_target)

  for longname in retdict:

    if retdict[longname][0]:
      goodlist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if goodlist:
    print "%4s %3s %25s %10s %30s" % ('ID','Own','Name','Status','Owner Information')

  # walk through target to print instead of the good list so that the
  # names are printed in order...
  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:

    if longname in goodlist:  
      if seash_global_variables.keys[environment_dict['currentkeyname']]['publickey'] == seash_global_variables.vesselinfo[longname]['ownerkey']:
        owner = '*'
      else:
        owner = ''
      print "%4s  %1s  %25s %10s %30s" % (seash_global_variables.vesselinfo[longname]['ID'],owner,seash_helper.fit_string(longname,25),seash_global_variables.vesselinfo[longname]['status'],seash_helper.fit_string(seash_global_variables.vesselinfo[longname]['ownerinfo'],30))

  # add groups for fail and good (if there is a difference in what nodes do)
  if goodlist and faillist:
    seash_global_variables.targets['listgood'] = goodlist
    seash_global_variables.targets['listfail'] = faillist
    print "Added group 'listgood' with "+str(len(seash_global_variables.targets['listgood']))+" targets and 'listfail' with "+str(len(seash_global_variables.targets['listfail']))+" targets"


  statusdict = {}
  # add status groups (if there is a difference in vessel state)
  for longname in goodlist:
    if seash_global_variables.vesselinfo[longname]['status'] not in statusdict:
      # create a list with this element...
      statusdict[seash_global_variables.vesselinfo[longname]['status']] = []
    statusdict[seash_global_variables.vesselinfo[longname]['status']].append(longname)

  if len(statusdict) > 1:
    print "Added group",
    for statusname in statusdict:
      seash_global_variables.targets['list'+statusname] = statusdict[statusname]
      print "'list"+statusname+"' with "+str(len(seash_global_variables.targets['list'+statusname]))+" targets",
    print
          

  
# reset                  -- Reset the vessel (clear the log and files and stop)
def reset(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  # reset the vessels...
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.reset_target)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  # and display it...
  seash_helper.print_vessel_errors(retdict)

  if goodlist and faillist:
    seash_global_variables.targets['resetgood'] = goodlist
    seash_global_variables.targets['resetfail'] = faillist
    print "Added group 'resetgood' with "+str(len(seash_global_variables.targets['resetgood']))+" targets and 'resetfail' with "+str(len(seash_global_variables.targets['resetfail']))+" targets"


  



# update
def update(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  # update information about the vessels...
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.list_or_update_target)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if goodlist and faillist:
    seash_global_variables.targets['updategood'] = goodlist
    seash_global_variables.targets['updatefail'] = faillist
    print "Added group 'updategood' with "+str(len(seash_global_variables.targets['updategood']))+" targets and 'updatefail' with "+str(len(seash_global_variables.targets['updatefail']))+" targets"

  


# savestate localfile -- Save current states to file (Added by Danny Y. Huang)
def savestate_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  # expand ~
  fileandpath = os.path.expanduser(command_key) 

  try:
    seash_helper.savestate(fileandpath, environment_dict['handleinfo'], 
                           environment_dict['host'], environment_dict['port'], 
                           environment_dict['expnum'], environment_dict['filename'], 
                           environment_dict['cmdargs'], environment_dict['defaulttarget'], 
                           environment_dict['defaultkeyname'], environment_dict['autosave'], 
                           environment_dict['currentkeyname'])
  except Exception, error:
    raise seash_exceptions.UserError("Error saving state: '" + str(error) + "'.")





# loadstate localfile -- Load states from file (Added by Danny Y. Huang)
def loadstate_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  # expand ~
  fileandpath = os.path.expanduser(command_key) 

  if not environment_dict['currentkeyname']:
    raise seash_exceptions.UserError("Specify the key name by first typing 'as [username]'.")

  # reading encrypted serialized states from file
  state_obj = open(fileandpath, 'r')
  cypher = state_obj.read()
  state_obj.close()

  try:
    # decrypt states
    statestr = rsa_decrypt(cypher, seash_global_variables.keys[environment_dict['currentkeyname']]['privatekey'])

    # deserialize
    state = serialize_deserializedata(statestr)
  except Exception, error:
    error_msg = "Unable to correctly parse state file. Your private "
    error_msg += "key may be incorrect."
    raise seash_exceptions.UserError(error_msg)

  # restore variables
  seash_global_variables.targets = state['targets']
  seash_global_variables.keys = state['keys']
  seash_global_variables.vesselinfo = state['vesselinfo']
  environment_dict['handleinfo'] = state['handleinfo']
  seash_global_variables.nextid = state['nextid']
  environment_dict['host'] = state['host']
  environment_dict['port'] = state['port']
  environment_dict['expnum'] = state['expnum']
  environment_dict['filename'] = state['filename']
  environment_dict['cmdargs'] = state['cmdargs'] 
  environment_dict['defaulttarget'] = state['defaulttarget']
  environment_dict['defaultkeyname'] = state['defaultkeyname']
  environment_dict['autosave'] = state['autosave']
  seash_global_variables.globalseashtimeout = state['globalseashtimeout']
  seash_global_variables.globaluploadrate = state['globaluploadrate']

  # Reload node handles. Rogue nodes are deleted from the original targets
  # and vesselinfo dictionaries.
  retdict = seash_helper.contact_targets(seash_global_variables.targets['%all'], seash_helper.reload_target, environment_dict['handleinfo'])

  reloadgood = []
  reloadfail = []

  for longname in retdict:
    if not retdict[longname][0]:
      reloadfail.append(longname)
    else:
      reloadgood.append(longname)
      
  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if reloadfail and reloadgood:
    seash_global_variables.targets['reloadgood'] = reloadgood
    seash_global_variables.targets['reloadfail'] = reloadfail
    print("Added group 'reloadgood' with " +str(len(seash_global_variables.targets['reloadgood'])) + \
            " targets and 'reloadfail' with " + str(len(seash_global_variables.targets['reloadfail'])) + " targets")

  if environment_dict['autosave']:
    print "Autosave is on."






# upload localfn (remotefn)   -- Upload a file 
def upload_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]


  # expand '~'
  fileandpath = os.path.expanduser(command_key)
  remotefn = os.path.basename(fileandpath)
  localfn = fileandpath


  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")


  # read the local file...
  fileobj = open(localfn,"r")
  filedata = fileobj.read()
  fileobj.close()


  # to prevent timeouts during file uploads on slow connections,
  # temporarily sets the timeout count on all vessels to be the
  # time needed to upload the file with the current globaluploadrate
  seash_helper.set_upload_timeout(filedata)


  # add the file to the vessels...
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.upload_target, remotefn, filedata)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['uploadgood'] = goodlist
    seash_global_variables.targets['uploadfail'] = faillist
    print "Added group 'uploadgood' with "+str(len(seash_global_variables.targets['uploadgood']))+" targets and 'uploadfail' with "+str(len(seash_global_variables.targets['uploadfail']))+" targets"


  # resets the timeout count on all vessels to globalseashtimeout
  seash_helper.reset_vessel_timeout()




# upload localfn remotefn
def upload_filename_remotefn(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  # expand '~'
  fileandpath = os.path.expanduser(command_key)
  localfn = fileandpath

  # Iterates through the dictionary to retrieve the user's argument input
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  remotefn = command_key


  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")


  # read the local file...
  fileobj = open(localfn,"r")
  filedata = fileobj.read()
  fileobj.close()


  # to prevent timeouts during file uploads on slow connections,
  # temporarily sets the timeout count on all vessels to be the
  # time needed to upload the file with the current globaluploadrate
  seash_helper.set_upload_timeout(filedata)


  # add the file to the vessels...
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.upload_target, remotefn, filedata)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['uploadgood'] = goodlist
    seash_global_variables.targets['uploadfail'] = faillist
    print "Added group 'uploadgood' with "+str(len(seash_global_variables.targets['uploadgood']))+" targets and 'uploadfail' with "+str(len(seash_global_variables.targets['uploadfail']))+" targets"


  # resets the timeout count on all vessels to globalseashtimeout
  seash_helper.reset_vessel_timeout()
  



# download remotefn (localfn) -- Download a file 
def download_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  # handle '~'
  fileandpath = os.path.expanduser(command_key)
  remotefn = os.path.basename(fileandpath)
  localfn = fileandpath

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")



  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.download_target,localfn,remotefn)

  writestring = ''

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
      # for output...
      writestring = writestring + retdict[longname][1]+ " "
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if writestring:
    print "Wrote files: "+writestring

  # add groups if needed...
  if goodlist and faillist:
    seash_global_variables.targets['downloadgood'] = goodlist
    seash_global_variables.targets['downloadfail'] = faillist
    print "Added group 'downloadgood' with "+str(len(seash_global_variables.targets['downloadgood']))+" targets and 'downloadfail' with "+str(len(seash_global_variables.targets['downloadfail']))+" targets"




# download remotefn localfn
def download_filename_localfn(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  remotefn = command_key

  # Iterates through the dictionary to retrieve the user's argument input
  while input_dict[command_key]['name'] is not 'args':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  # handle '~'
  fileandpath = os.path.expanduser(command_key)
  localfn = fileandpath

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")



  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.download_target,localfn,remotefn)

  writestring = ''

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
      # for output...
      writestring = writestring + retdict[longname][1]+ " "
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if writestring:
    print "Wrote files: "+writestring

  # add groups if needed...
  if goodlist and faillist:
    seash_global_variables.targets['downloadgood'] = goodlist
    seash_global_variables.targets['downloadfail'] = faillist
    print "Added group 'downloadgood' with "+str(len(seash_global_variables.targets['downloadgood']))+" targets and 'downloadfail' with "+str(len(seash_global_variables.targets['downloadfail']))+" targets"
  
  



# delete remotefn             -- Delete a file
def delete_remotefn(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  remotefn = command_key

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")


  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.delete_target, remotefn)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else: 
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # add groups if needed...
  if goodlist and faillist:
    seash_global_variables.targets['deletegood'] = goodlist
    seash_global_variables.targets['deletefail'] = faillist
    print "Added group 'deletegood' with "+str(len(seash_global_variables.targets['deletegood']))+" targets and 'deletefail' with "+str(len(seash_global_variables.targets['deletefail']))+" targets"




# cat remotefn  -- Print a file to the screen
def cat_filename(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  remotefn = os.path.basename(command_key)

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")



  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.cat_target,remotefn)


  # here is where I list it out...
  for longname in retdict:

    # True means it worked
    if retdict[longname][0]:
      print
      print "File '"+remotefn+"' on '"+longname+"': "
      print retdict[longname][1]
      print
      goodlist.append(longname)

    else:
      print
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  if goodlist and faillist:
    seash_global_variables.targets['catgood'] = goodlist
    seash_global_variables.targets['catfail'] = faillist
    print "Added group 'catgood' with "+str(len(seash_global_variables.targets['catgood']))+" targets and 'catfail' with "+str(len(seash_global_variables.targets['catfail']))+" targets"




  
  

  
# start file [args ...]  -- Start an experiment
def start_remotefn(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the repy version to run
  while input_dict[command_key]['name'] is not 'start':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  command = command_key

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  argstring = ' '.join([command_key])

  prog_platform = seash_helper.get_execution_platform(command,
    command_key)

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  # need to get the status, etc (or do I just try to start them all?)
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(
    seash_global_variables.targets[environment_dict['currenttarget']],
    seash_helper.start_target, argstring, prog_platform)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # add groups if needed...
  if goodlist and faillist:
    seash_global_variables.targets['startgood'] = goodlist
    seash_global_variables.targets['startfail'] = faillist
    print "Added group 'startgood' with "+str(len(seash_global_variables.targets['startgood']))+" targets and 'startfail' with "+str(len(seash_global_variables.targets['startfail']))+" targets"





# start file arg
def start_remotefn_arg(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the repy version to run
  while input_dict[command_key]['name'] is not 'start':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  command = command_key

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  filename = command_key

  # Iterates through the dictionary to retrieve the user's argument input
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  arguments = command_key

  argstring = ' '.join([filename] + [arguments])

  prog_platform = seash_helper.get_execution_platform(command,
    filename)

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  # need to get the status, etc (or do I just try to start them all?)
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(
    seash_global_variables.targets[environment_dict['currenttarget']],
    seash_helper.start_target, argstring, prog_platform)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # add groups if needed...
  if goodlist and faillist:
    seash_global_variables.targets['startgood'] = goodlist
    seash_global_variables.targets['startfail'] = faillist
    print "Added group 'startgood' with "+str(len(seash_global_variables.targets['startgood']))+" targets and 'startfail' with "+str(len(seash_global_variables.targets['startfail']))+" targets"





# stop               -- Stop an experiment
def stop(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  # need to get the status, etc (or do I just try to stop them all?)
  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.stop_target)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)


  # add groups if needed...
  if goodlist and faillist:
    seash_global_variables.targets['stopgood'] = goodlist
    seash_global_variables.targets['stopfail'] = faillist
    print "Added group 'stopgood' with "+str(len(seash_global_variables.targets['stopgood']))+" targets and 'stopfail' with "+str(len(seash_global_variables.targets['stopfail']))+" targets"




# run file [args...]    -- Shortcut for upload a file and start
def run_localfn(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the repy version to run
  while input_dict[command_key]['name'] is not 'run':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  command = command_key

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
     input_dict = input_dict[command_key]['children']
     command_key = input_dict.keys()[0]

  # Handle '~' in file names
  fileandpath = os.path.expanduser(command_key)
  onlyfilename = os.path.basename(fileandpath)

  argstring = " ".join([onlyfilename])

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  prog_platform = seash_helper.get_execution_platform(command,
    onlyfilename)

  # read the local file...
  fileobj = open(fileandpath,"r")
  filedata = fileobj.read()
  fileobj.close()


  # to prevent timeouts during file uploads on slow connections, 
  # temporarily sets the timeout count on all vessels to be the  
  # time needed upload the file with the current globaluploadrate 
  seash_helper.set_upload_timeout(filedata) 


  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(
    seash_global_variables.targets[environment_dict['currenttarget']],
    seash_helper.run_target, onlyfilename, filedata, argstring,
    prog_platform)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['rungood'] = goodlist
    seash_global_variables.targets['runfail'] = faillist
    print "Added group 'rungood' with "+str(len(seash_global_variables.targets['rungood']))+" targets and 'runfail' with "+str(len(seash_global_variables.targets['runfail']))+" targets"


  # resets the timeout count on all vessels to globalseashtimeout 
  seash_helper.reset_vessel_timeout() 





# run file arg
def run_localfn_arg(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the repy version to run
  while input_dict[command_key]['name'] is not 'run':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  command = command_key

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  # Handle '~' in file names
  fileandpath = os.path.expanduser(command_key)
  onlyfilename = os.path.basename(fileandpath)

  # Iterates down one more level to retrieve the argument string
  input_dict = input_dict[command_key]['children']
  command_key = input_dict.keys()[0]

  argstring = " ".join([onlyfilename] + [command_key])

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  prog_platform = seash_helper.get_execution_platform(command,
    onlyfilename)

  # read the local file...
  fileobj = open(fileandpath,"r")
  filedata = fileobj.read()
  fileobj.close()


  # to prevent timeouts during file uploads on slow connections, 
  # temporarily sets the timeout count on all vessels to be the  
  # time needed upload the file with the current globaluploadrate 
  seash_helper.set_upload_timeout(filedata) 


  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(
    seash_global_variables.targets[environment_dict['currenttarget']],
    seash_helper.run_target, onlyfilename, filedata, argstring,
    prog_platform)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['rungood'] = goodlist
    seash_global_variables.targets['runfail'] = faillist
    print "Added group 'rungood' with "+str(len(seash_global_variables.targets['rungood']))+" targets and 'runfail' with "+str(len(seash_global_variables.targets['runfail']))+" targets"


  # resets the timeout count on all vessels to globalseashtimeout 
  seash_helper.reset_vessel_timeout() 




#split resourcefn            -- Split off of each vessel another vessel
def split_resourcefn(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's filename
  while input_dict[command_key]['name'] is not 'filename':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  # expand ~
  fileandpath = os.path.expanduser(command_key)
  resourcefn = fileandpath

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  resourcefo = open(resourcefn)
  resourcedata = resourcefo.read()
  resourcefo.close() 


  faillist = []
  goodlist1 = []
  goodlist2 = []

  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.split_target,resourcedata)

  for longname in retdict:
    if retdict[longname][0]:
      newname1, newname2 = retdict[longname][1]
      goodlist1.append(newname1)
      goodlist2.append(newname2)
      print longname+" -> ("+newname1+", "+newname2+")"
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist1 and goodlist2 and faillist:
    seash_global_variables.targets['split1'] = goodlist1
    seash_global_variables.targets['split2'] = goodlist2
    seash_global_variables.targets['splitfail'] = faillist
    print "Added group 'split1' with "+str(len(seash_global_variables.targets['split1']))+" targets, 'split2' with "+str(len(seash_global_variables.targets['split2']))+" targets and 'splitfail' with "+str(len(seash_global_variables.targets['splitfail']))+" targets"
  elif goodlist1 and goodlist2:
    seash_global_variables.targets['split1'] = goodlist1
    seash_global_variables.targets['split2'] = goodlist2
    print "Added group 'split1' with "+str(len(seash_global_variables.targets['split1']))+" targets and 'split2' with "+str(len(seash_global_variables.targets['split2']))+" targets"

  


#join                        -- Join vessels on the same node
def join(input_dict, environment_dict):

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  if not environment_dict['currentkeyname'] or not seash_global_variables.keys[environment_dict['currentkeyname']]['publickey'] or not seash_global_variables.keys[environment_dict['currentkeyname']]['privatekey']:
    raise seash_exceptions.UserError("Must specify an identity with public and private keys...")

  nodedict = {}
  skipstring = ''

  # Need to group vessels by node...
  for longname in seash_global_variables.targets[environment_dict['currenttarget']]:
    if seash_global_variables.keys[environment_dict['currentkeyname']]['publickey'] != seash_global_variables.vesselinfo[longname]['ownerkey']:
      skipstring = skipstring + longname+" "
      continue

    nodename = seash_global_variables.vesselinfo[longname]['IP']+":"+str(seash_global_variables.vesselinfo[longname]['port'])

    if nodename not in nodedict:
      nodedict[nodename] = []

    nodedict[nodename].append(longname)

  # if we skip nodes, explain why
  if skipstring:
    print "Skipping "+skipstring+" because the current identity is not the owner."
    print "If you are trying to join vessels with different owners, you need"
    print "to change ownership to the same owner first"


  faillist = []
  goodlist = []

  retdict = seash_helper.contact_targets(nodedict.keys(),seash_helper.join_target,nodedict)

  for nodename in retdict:

    if retdict[nodename][0]:
      print retdict[nodename][1][0],"<- ("+", ".join(nodedict[nodename])+")"
      goodlist = goodlist + nodedict[nodename]
    else:
      if retdict[nodename][1]:
        faillist = faillist + nodedict[nodename]
      # Nodes that I only have one vessel on don't get added to a list...
  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['joingood'] = goodlist
    seash_global_variables.targets['joinfail'] = faillist
    print "Added group 'joingood' with "+str(len(seash_global_variables.targets['joingood']))+" targets and 'joinfail' with "+str(len(seash_global_variables.targets['joinfail']))+" targets"
  elif goodlist:
    seash_global_variables.targets['joingood'] = goodlist
    seash_global_variables.targets['joinfail'] = faillist
    print "Added group 'joingood' with "+str(len(seash_global_variables.targets['joingood']))+" targets"




###Set Commands###

# set                 -- Changes the state of the targets (use 'help set')
def set(input_dict, environment_dict):
  pass
        




# set owner identity        -- Change a vessel's owner
def set_owner_arg(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  newowner = command_key

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  if newowner not in seash_global_variables.keys:
    raise seash_exceptions.UserError("Unknown identity: '"+newowner+"'")

  if not seash_global_variables.keys[newowner]['publickey']:
    raise seash_exceptions.UserError("No public key for '"+newowner+"'")

  faillist = []
  goodlist = []
  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.setowner_target,newowner)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)


  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['ownergood'] = goodlist
    seash_global_variables.targets['ownerfail'] = faillist
    print "Added group 'ownergood' with "+str(len(seash_global_variables.targets['ownergood']))+" targets and 'ownerfail' with "+str(len(seash_global_variables.targets['ownerfail']))+" targets"
  




# set autosave [ on | off ] -- Set whether SeaSH automatically saves the last state.
def set_autosave_arg(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  if command_key == 'on':
    environment_dict['autosave'] = True
  elif command_key == 'off':
    environment_dict['autosave'] = False
  else:
    raise seash_exceptions.UserError("Usage: set autosave [ on | off (default)]")





# set advertise [ on | off ] -- Change advertisement of vessels
def set_advertise_arg(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  if command_key == 'on':
    newadvert = True
  elif command_key == 'off':
    newadvert = False
  else:
    raise seash_exceptions.UserError("Usage: set advertise [ on | off ]")

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")


  faillist = []
  goodlist = []
  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.setadvertise_target,newadvert)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['advertisegood'] = goodlist
    seash_global_variables.targets['advertisefail'] = faillist
    print "Added group 'advertisegood' with "+str(len(seash_global_variables.targets['advertisegood']))+" targets and 'advertisefail' with "+str(len(seash_global_variables.targets['advertisefail']))+" targets"
  
    

  

# set ownerinfo [ data ... ]    -- Change owner information for the vessels
def set_ownerinfo_arg(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  newdata = command_key

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")

  faillist = []
  goodlist = []
  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.setownerinformation_target,newdata)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)
  seash_helper.print_vessel_errors(retdict)


  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['ownerinfogood'] = goodlist
    seash_global_variables.targets['ownerinfofail'] = faillist
    print "Added group 'ownerinfogood' with "+str(len(seash_global_variables.targets['ownerinfogood']))+" targets and 'ownerinfofail' with "+str(len(seash_global_variables.targets['ownerinfofail']))+" targets"


  

# set users [ identity ... ]  -- Change a vessel's users
def set_users_arg(input_dict, environment_dict):
  userkeys = []

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  userlist = command_key.split()

  for identity in userlist:
    if identity not in seash_global_variables.keys:
      raise seash_exceptions.UserError("Unknown identity: '"+identity+"'")

    if not seash_global_variables.keys[identity]['publickey']:
      raise seash_exceptions.UserError("No public key for '"+identity+"'")

    userkeys.append(rsa_publickey_to_string(seash_global_variables.keys[identity]['publickey']))
  # this is the format the NM expects...
  userkeystring = '|'.join(userkeys)

  if not environment_dict['currenttarget']:
    raise seash_exceptions.UserError("Must specify a target")


  faillist = []
  goodlist = []
  retdict = seash_helper.contact_targets(seash_global_variables.targets[environment_dict['currenttarget']],seash_helper.setusers_target,userkeystring)

  for longname in retdict:
    if retdict[longname][0]:
      goodlist.append(longname)
    else:
      faillist.append(longname)

  seash_helper.print_vessel_errors(retdict)

  # update the groups
  if goodlist and faillist:
    seash_global_variables.targets['usersgood'] = goodlist
    seash_global_variables.targets['usersfail'] = faillist
    print "Added group 'usersgood' with "+str(len(seash_global_variables.targets['usersgood']))+" targets and 'usersfail' with "+str(len(seash_global_variables.targets['usersfail']))+" targets"
  
    


# set timeout count  -- Sets the time that seash is willing to wait on a node
def set_timeout_arg(input_dict, environment_dict):

  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  # I need to set the timeout for new handles...
  try:
    seash_global_variables.globalseashtimeout = int(command_key)
  except ValueError:
    raise seash_exceptions.UserError("The timeout value must be a number")

  # let's reset the timeout for existing handles...
  for longname in seash_global_variables.vesselinfo:
    thisvesselhandle = seash_global_variables.vesselinfo[longname]['handle']
    thisvesselhandledict = nmclient_get_handle_info(thisvesselhandle)
    thisvesselhandledict['timeout'] = seash_global_variables.globalseashtimeout
    nmclient_set_handle_info(thisvesselhandle,thisvesselhandledict)


          
# set uploadrate speed     -- Sets the speed seash will modify the timeout count with when uploading files
def set_uploadrate_arg(input_dict, environment_dict):

  try:
    command_key = input_dict.keys()[0]

    # Iterates through the dictionary to retrieve the user's argument
    while input_dict[command_key]['name'] is not 'args':
      input_dict = input_dict[command_key]['children']
      command_key = input_dict.keys()[0]

    seash_global_variables.globaluploadrate = int(command_key)
  except ValueError:
    raise seash_exceptions.UserError("The speed value must be a number (in bytes per second)")


# set showparse [on | off] -- Toggles display of the parsed commandline input
def set_showparse_args(input_dict, environment_dict):
  command_key = input_dict.keys()[0]

  # Iterates through the dictionary to retrieve the user's argument
  while input_dict[command_key]['name'] is not 'args':
    input_dict = input_dict[command_key]['children']
    command_key = input_dict.keys()[0]

  command_key = command_key.lower()
  if command_key == 'on':
    environment_dict['showparse'] = True
  elif command_key == 'off':
    environment_dict['showparse'] = False
  else:
    raise seash_exceptions.UserError("Value can be either 'on' or 'off'") 


# enable modulename
def enable_module(input_dict, environment_dict):
  # Enables an installed module.

  # Get the modulename
  dict_mark = input_dict
  try:
    command = dict_mark.keys()[0]
    while dict_mark[command]['name'] != 'modulename':
      dict_mark = input_dict[command]['children']
      command = dict_mark.keys()[0]
    modulename = command
  except IndexError:
    raise seash_exceptions.UserError("Error, command requires a modulename")

  try:
    seash_modules.enable(seash_dictionary.seashcommanddict, modulename)
  except seash_exceptions.ModuleConflictError, e:
    print "Module cannot be imported due to the following conflicting command:"
    print str(e)


# disable modulename
def disable_module(input_dict, environment_dict):
  # Disables an enabled module.

  # Get the modulename
  dict_mark = input_dict
  try:
    command = dict_mark.keys()[0]
    while dict_mark[command]['name'] != 'modulename':
      dict_mark = input_dict[command]['children']
      command = dict_mark.keys()[0]
    modulename = command
  except IndexError:
    raise seash_exceptions.UserError("Error, command requires a modulename")

  seash_modules.disable(seash_dictionary.seashcommanddict, modulename)

# modulehelp modulename
def print_module_help(input_dict, environment_dict):
  # Prints the helptext for a module

  # Get the modulename
  dict_mark = input_dict
  try:
    command = dict_mark.keys()[0]
    while dict_mark[command]['name'] != 'modulename':
      dict_mark = input_dict[command]['children']
      command = dict_mark.keys()[0]
    modulename = command
  except IndexError:
    raise seash_exceptions.UserError("Error, command requires a modulename")

  # Is this module installed?
  if not modulename in seash_modules.module_data:
    raise seash_exceptions.UserError("Module is not installed.")

  print seash_modules.module_data[modulename]['help_text']

  # Now, print out all the commands under this module
  print "Commands in this module:"
  print '\n'.join(seash_helper.get_commands_from_commanddict(seash_modules.module_data[modulename]['command_dict']))
  


# show modules
def list_all_modules(input_dict, environment_dict):
  # Lists all modules and their status.
  print "Enabled Modules:"
  print ", ".join(seash_modules.get_enabled_modules())
  print
  print "Installed Modules:"

  # Now print the URLs...
  # Output format:
  # modulename - URL not available  # URL is set to None
  # modulename - https://seattle.poly.edu/seashplugins/modulename/ # URL is set
  for module in seash_modules.module_data:
    print module, '-',
    if seash_modules.module_data[module]['url'] is None:
      print "Install URL not available"
    else:
      print seash_modules.module_data[module]['url']
  print



# exit
def exit(input_dict, environment_dict):
  sys.exit()



"""
Author: Alan Loh, Danny Huang

Module: A class that allows for tab completion of command words and certain user
        arguments in the seash command prompt. Available completions are 
        determined by the seash command dictionary.

Upon initialization of the seash command prompt, a Completer object is
initialized and maintains a copy of the seash command dictionary. At the beginning
of each command loop, the Completer will update its arguments list for the
purpose of tab completion for target IDs and loaded keynames.

When the user double tabs, seash passes the command inputted thus far to the tab
completer. Tab completion first determines where in the command dictionary has
the user inputted up to thus far, and a list of possible command completions is 
built based on the children of the last command string inputted. A
list of commands will be returned if there are multiple possible completions, or
the tab completer will automatically complete the user input in the command
prompt if there's only one completion available.

There is a slight issue of inefficiency in that each time tab completion is
called with a new user prefix, the tab completer will always include the user's
prefix in addition to the completed word as part of the returned completion.
Since seash commands are generally short in length, this is not an issue yet,
but there is still room for improvement.

This only applies for operating systems that supports readline. This particular
version's file name completer is compatible only with systems that uses '/' in
their directory paths.
"""

import os
import os.path
import seash_dictionary
import seash_global_variables
import seash_modules

class Completer:
  def __init__(self):
    self._prefix = None
    # Retrieves a copy of seash's command dictionary for reference
    self.commanddict = seash_dictionary.return_command_dictionary()




  # Creates and updates the list of targets and groups for tab completion
  def set_target_list(self):
    self.targetList = seash_global_variables.targets.keys()[:]




  # Creates and updates the list of available key names for tab completion
  def set_keyname_list(self):
    self.keynameList = seash_global_variables.keys.keys()



  # Returns the path from a given prefix, by extracting the string up to the
  # last forward slash in the prefix. If no forward slash is found, returns an
  # empty string.
  def _getpath(self, prefix):

    slashpos = prefix.rfind("/")
    currentpath = ""
    if slashpos > -1:
      currentpath = prefix[0 : slashpos+1]

    return currentpath



  # Returns the file name, or a part of the file name, from a given prefix, by
  # extracting the string after the last forward slash in the prefix. If no
  # forward slash is found, returns an empty string.
  def _getfilename(self, prefix):

    # Find the last occurrence of the slash (if any), as it separates the path
    # and the file name.
    slashpos = prefix.rfind("/")
    filename = ""

    # If slash exists and there are characters after the last slash, then the
    # file name is whatever that follows the last slash.
    if slashpos > -1 and slashpos+1 <= len(prefix)-1:
      filename = prefix[slashpos+1:]

    # If no slash is found, then we assume that the entire user input is the
    # prefix of a file name because it does not contain a directory
    elif slashpos == -1:
      filename = prefix

    # If both cases fail, then the entire user input is the name of a
    # directory. Thus, we return the file name as an empty string.
    return filename




  # Returns a list of file names that start with the given prefix.
  def _listfiles(self, prefix):
    
    # Find the directory specified by the prefix
    currentpath = self._getpath(prefix)
    if not currentpath:
      currentpath = "./"
    filelist = []

    # Attempt to list files from the given directory
    try:
      currentpath = os.path.expanduser(currentpath)
      filelist = os.listdir(currentpath)

    # In the case an exception occurs, return filelist anyway as it will be empty
    finally:
      return filelist




  # Iterates through the command dictionary and returns the list of children of the last valid 
  # command in the passed string list
  def _get_all_commands(self, input_list):
    # The iterator that will iterate through the command dictionary based on what
    # the user has inputted thus far
    dict_iterator = self.commanddict

    # The list of command completions that will be built throughout the method
    completion_list = []

    # Reference to keep track of the last string the user inputted
    last_string = ""

    # Determines if the last string the user inputted was incomplete
    incomplete_string = False


    # Iterate through the input list to determine current path down the command dictionaries
    for commands in input_list:
      last_string = commands

      if commands in dict_iterator.keys():
        dict_iterator = dict_iterator[commands]['children']


      # Test the possibility of an user argument in the input list by first
      # seeing if the command exists as one of the child's of the current key,
      # and, if applicable, the user's input corresponds to one of the possibilities
      # in the list of possible argument names
      elif '[TARGET]' in dict_iterator.keys() and commands in self.targetList:
        dict_iterator = dict_iterator['[TARGET]']['children']
      elif '[GROUP]' in dict_iterator.keys() and commands in self.targetList:
        dict_iterator = dict_iterator['[GROUP]']['children']
      elif '[KEYNAME]' in dict_iterator.keys() and commands in self.keynameList:
        dict_iterator = dict_iterator['[KEYNAME]']['children']
      elif '[FILENAME]' in dict_iterator.keys() and self._getfilename(commands) in self._listfiles(commands):
        dict_iterator = dict_iterator['[FILENAME]']['children']
      elif '[ARGUMENT]' in dict_iterator.keys():
        dict_iterator = dict_iterator['[ARGUMENT]']['children']


      # Otherwise, if it's not an user argument nor is it a command word, set
      # incomplete_string to be true
      else:
        incomplete_string = True


    # Reconstructs the user's input up to the last string if it's incomplete,
    # or just reconstructs the user's input entirely if that's not the case
    if incomplete_string:
      user_prefix = " ".join(input_list[:input_list.index(last_string)])
    else:
      user_prefix = " ".join(input_list[:])


    # Creates a list of all the possible command completions 
    for child in dict_iterator.keys():

      # User argument completion. All argument key names begins with '['
      if child.startswith('['):
        # Tab completer will not differentiate between a group from a target
        if child == '[TARGET]' or child == '[GROUP]':
          for targets in self.targetList:
            completed_command = user_prefix + ' ' + targets
            completion_list.append(completed_command.strip() + ' ')
      
        elif child == '[KEYNAME]':
          for keynames in self.keynameList:
            completed_command = user_prefix + ' ' + keynames
            completion_list.append(completed_command.strip() + ' ')
      
        elif child == '[FILENAME]':
          # Find the directory specified by the user's last inputted string
          currentpath = self._getpath(last_string)
          if not currentpath:
            currentpath = "./"

          # Finds the full path of the user's given directory in case a '~' was given
          currentpath = os.path.expanduser(currentpath)

          for filenames in self._listfiles(last_string):
            dirfile = os.path.join(currentpath, filenames)
            completed_command = user_prefix + ' ' + self._getpath(last_string) + filenames

            # For directory names, adds a forward slash at the end of the tab completion
            # For file names, adds a space at the end of the tab completion
            if os.path.isdir(dirfile):
              completion_list.append(completed_command.strip() + '/')
            else:
              completion_list.append(completed_command.strip() + ' ')
      

      # Command word completion
      else:
        completed_command = user_prefix + ' ' + child
        completion_list.append(completed_command.strip() + ' ')

    return completion_list





  # The actual tab completion method
  def complete(self, prefix, index):
    # If the passed prefix is different than the previously passed prefix,
    # then a new list of possible completions needs to be constructed corresponding
    # to the new prefix
    if prefix != self._prefix:
      self._matching_words = []

      # split the user input into a list of strings
      input_list = prefix.split()
      # Retrieves the list of children of the current command dictionary based on
      # the current input of the user
      self._words = self._get_all_commands(input_list)
      
      # Retrieves the list of words from modules based on the current input of 
      # the user
      self._words += seash_modules.tab_complete(input_list)
      
      for word in self._words:
        if word.startswith(prefix):
          self._matching_words.append(word)

      # updates prefix to determine when a new completion list needs to be made
      self._prefix = prefix
    

    try:
      return self._matching_words[index]
    except IndexError:
      return None
